(*
 *
 * Copyright (c) 2004, 
 *  Jeremy Condit       <jcondit@cs.berkeley.edu>
 *  George C. Necula    <necula@cs.berkeley.edu>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

open Cil
open Pretty
open Dattrs
open Doptions
open Dutil

module E = Errormsg
module GA = GrowArray
module IH = Inthash
module RD = Reachingdefs
module AE = Availexps
module RCT = Rmciltmps
module DF = Dataflow
module UD = Usedef
module S = Stats


let (<%) = (fun x y -> (Int64.compare x y) < 0)
let (<=%) = (fun x y -> (Int64.compare x y) <= 0)
let (>%) = (fun x y -> (Int64.compare x y) > 0)
let (>=%) = (fun x y -> (Int64.compare x y) >= 0)
let (<>%) = (fun x y -> (Int64.compare x y) <> 0)
   
let (+%) = Int64.add
let (-%) = Int64.sub
let ( *% ) = Int64.mul
let (/%) = Int64.div
let (~-%) = Int64.neg

(* let int64to32 (i: int64) : int32 =  *)
(*   let i' = Int64.to_int32 i in (\* Silently drop the high 32 bits *\) *)
(*   if i = Int64.of_int32 i' then i' *)
(*   else E.s (unimp "A constant that doesn't fit in 32 bits.") *)

(* What is the largest number that can be stored in the given integral type,
   assuming it is treated as unsigned? *)
let maxUnsignedInt (t:typ) : int64 =
  (Int64.shift_left 1L (bitsSizeOf t)) -% 1L

let debugOptim = false

(************************************************************************)


let rec isIntOrPtrType (t:typ) : bool =
  match t with
    TInt _ | TPtr _ -> true
  | TNamed (tt,_) -> isIntOrPtrType tt.ttype
  | _ -> false

(* Is this a signed comparison? *)
let rec isSignedType (t:typ) : bool =
  match t with
    TInt(ik,_) -> isSigned ik
  | TPtr _ -> false
  | TNamed (tt, _) -> isSignedType tt.ttype
  | _ -> E.s (bug "expecting an int or ptr in isSignedType.")

(* Returns the size of a pointer's base type in bytes, if known *)
let sizeOfBaseType ptrt: int option =
  match unrollType ptrt with
  | TPtr (bt, _) -> begin
      match isInteger (constFold true (SizeOf bt)) with
      | Some n -> Some (to_int n)
      | None -> None
    end
  | _ -> (* maybe the expression is NULL *)
      None

(* Do we need an alignment check for p + x?  Well, that depends on the size of
 *  *p.  If the size is a power of two, p + x will be aligned even if it
 *  overflows, so we can skip the check. *)
let needsAlignCheck ptrt: bool =
  match sizeOfBaseType ptrt with (* Look for common multiples of 2 *)
    Some (1|2|4|8|16|32|64|128|256|512|1024|2048|4096) -> false
  | _ -> true


(** Keep expressions to be compared in a canonical form: a constant + sum of 
 * weighted expressions, where the latter are not something that can be 
 * broken in a canonical expression themselves. These atomic expressions will 
 * be compared for equality.  *)
module Can = struct
  type t = 
      { ct: int;
        cf: (int * exp) list;
      }
  let mkInt n = { ct = n; cf = [] }
  let atomic (f: int) (e: exp) = { ct = 0; cf = [(f, e)]}

  let zero  = mkInt 0
  let weightedAdd (w1: int) (c1: t) (cacc: t) = 
    let rec insert (w: int) (e: exp) = function
        [] -> [ (w, e) ]
      | (w', e') :: rest -> 
          if compareExp e e' then 
            let w'' = w + w' in 
            if w'' = 0 then 
              rest
            else
              (w'', e') :: rest
          else
            (w', e') :: insert w e rest
    in
    { ct = w1 * c1.ct + cacc.ct;
      cf = List.fold_left (fun acc (w,e) -> insert (w1 * w) e acc) 
                          cacc.cf c1.cf
    }

  let add c1 c2 = weightedAdd 1 c1 c2
  let sub c1 c2 = weightedAdd (-1) c2 c1
  let mult c n  = weightedAdd n c zero

  let addConst cnst c = { c with ct = cnst + c.ct}

  let d_t () c =
    if c.cf = [] then 
      num c.ct
    else begin
      dprintf "%a%a" 
        insert
        (if c.ct = 0 then nil else dprintf "%d" c.ct)
        (docList ~sep:nil
           (fun (w, e) -> 
             dprintf "%a%a" 
               insert
               (if w = 1 then chr '+' 
               else if w = -1 then chr '-'
               else 
                 (if w > 0 then chr '+' else nil) ++ 
                   dprintf "%d*" w)
               d_exp e))
        c.cf
    end
end

(** The arithmetic factor for a base type *)
let arithFactor (t: typ) : int = 
  match unrollType t with 
    TPtr (bt, _)  -> begin
      let bs = try bitsSizeOf bt with _ -> E.s (error "Pointer arithmetic on unknown base type") in
      bs / 8
    end
  | _ -> 1
  
(** Convert an expression into a canonical expression *)
let rec canonExp (fact: int) (e: exp) : Can.t = 
  match e with 
    Const (CInt64 (i, _, _)) -> Can.mkInt (fact * (Int64.to_int i))
  | BinOp ((PlusA|PlusPI|IndexPI), e1, e2, t) -> 
      let facte2 : int = fact * arithFactor t in
      Can.add (canonExp fact e1) (canonExp facte2 e2)

  | BinOp ((MinusA|MinusPI), e1, e2, t) -> 
      let facte2 : int = fact * arithFactor t in
      Can.add (canonExp fact e1) (canonExp (- facte2) e2)
      
  | CastE _ -> begin
      let e' = stripNopCasts e in 
      if e' != e then 
        canonExp fact e'
      else
        Can.atomic fact e'
  end

  | _ -> Can.atomic fact e

let canonExp (fact: int) (e: exp) : Can.t = 
  let e = constFold true e in (* Apply constant folding first *)
  if debugOptim then begin
    let res = canonExp fact e in 
    ignore (log "canonExp(%a) = %a\n" d_exp e Can.d_t res);
    res
  end else
    canonExp fact e


(** Split an expression e into e',i', such that e=e'+i' (pointer arithmetic) *)
let rec getBaseOffset (e: exp) : exp * int =
  match e with
  | BinOp ((PlusPI | IndexPI | MinusPI) as op, e', off, t) ->
      let intFold e = isInteger (constFold true e) in
      begin
        match getBaseOffset e', intFold off, op with
        | (b, n1), Some n2, (PlusPI | IndexPI) ->
            b, n1 + (to_signedint n2)
        | (b, n1), Some n2, MinusPI ->
            b, n1 - (to_signedint n2)
        | _, _, _ ->
            e, 0
      end
  | CastE (t, e') ->
      let strip t = typeRemoveAttributes ["bounds"; "size"; "nullterm"] t in
      let b, n = getBaseOffset e' in
      if compareTypes (strip (typeOf e')) (strip t) then begin
        b, n
      end else begin
        match sizeOfBaseType (typeOf e'), sizeOfBaseType t with
        | Some sz1, Some sz2 when sz1 = sz2 -> b, n
        | Some sz1, Some sz2 when sz1 mod sz2 = 0 ->
            mkCast b t, n * (sz1 / sz2)
        | _ -> e, 0
      end
  | AddrOf (base, off) -> begin (* Look for when off ends in Index *)
      match removeOffset off with 
       off', Index(idx, NoOffset) -> begin
         match isInteger (constFold true idx) with 
           Some n -> mkAddrOrStartOf (base, off'), (to_signedint n)
         | _ -> e, 0
       end
      | _ -> e, 0
  end
      
  | _ ->
      e, 0

let proveLeWithBounds (e1: exp) (e2: exp) : bool =
  let ctx = allContext () in
  let rec getExpBounds (e:exp) : exp option * exp option =
    match e with
      (* TODO: structs, memory *)
    | Lval (Var vi, NoOffset) when isPointerType vi.vtype &&
                                   hasAttribute "bounds" (typeAttrs vi.vtype) ->
        let ctx = addThisBinding ctx vi.vtype e in
        let lo, hi = boundsOfAttrs ctx (typeAttrs vi.vtype) in
(*         log " %a has bounds %a and %a.\n" dx_exp e *)
(*           dx_exp lo dx_exp hi; *)
        Some lo, Some hi
    | CastE (t, e') when (bitsSizeOf t) = (bitsSizeOf voidPtrType) ->
        getExpBounds e'
    | _ -> None, None
  in
  let lo1, hi1 = getExpBounds e1 in
  let lo2, hi2 = getExpBounds e2 in
  (* we know e1 <= hi1 and lo2 <= e2 *)
  match hi1, lo2 with
    Some hi1, Some lo2 ->
      (compareExpStripCasts hi1 lo2)
      || (compareExpStripCasts hi1 e2)
      || (compareExpStripCasts e1 lo2)
  | Some hi1, None ->
      (compareExpStripCasts hi1 e2)
  | None, Some lo2 ->
      (compareExpStripCasts e1 lo2)
  | None, None ->
      false

(** Return true if e1 <= e2 (statically). Reports an error if we can prove 
 * statically e1 > e2, unless allowGt = true *)
let proveLe ?(allowGt: bool = false) (e1: exp) (e2: exp) (why: string) : bool =
  let b1, off1 = getBaseOffset (stripNopCasts e1) in
  let b2, off2 = getBaseOffset (stripNopCasts e2) in
  if false then begin
    log "  proveLeq: Comparing:\n"; 
    log "   %a = (%a) + %d.\n" dx_exp e1 d_plainexp b1 off1; 
    log "   %a = (%a) + %d.\n" dx_exp e2 d_plainexp b2 off2; 
  end;
  if compareExpStripCasts b1 b2 then begin
    let doCompare n1 n2 =
      if n1 > n2 then begin
        if not allowGt then
          error "Bounds check (%s) will always fail: %a <= %a" why dx_exp e1 dx_exp e2;
        false
      end else begin
        true
      end
    in
    let t1 = typeOf b1 in
    let t2 = typeOf b2 in
    match sizeOfBaseType t1, sizeOfBaseType t2 with
    | Some n1, Some n2 -> doCompare (off1 * n1) (off2 * n2)
    | _ when compareTypes t1 t2 -> doCompare off1 off2
    | _ -> 
(*         log "   Comparetypes %a, %a failed.\n" dx_type t1 dx_type t2; *)
        false
  end else begin
    (proveLeWithBounds b1 b2 && off1 = 0 && off2 = 0)
  end

let newProveLe ?(allowGt: bool = false) (e1: exp) (e2: exp) : bool =
  let e1c = canonExp 1 e1 in
  let e2c = canonExp 1 e2 in 
  let e1res = Can.sub e1c e2c in 
  if debugOptim then 
    log "le(%a, %a) = %a\n" d_exp e1 d_exp e2 Can.d_t e1res;
  false

(** Optimize an individual check *)
let rec optimizeCheck (c: check) : check list =
  if false then 
    ignore (log "optimizing %a\n"
              d_instr (checkToInstr c)); 
  match c with
  | CNullOrLeq (z, _, _, _) when isZero z -> []

  | CNullOrLeq (_, e1, e2, why)
  | CLeq (e1, e2, why) ->
      (* If e1 is statically larger than e2, this test could still pass
         if e == 0.  But Report an error, since it's probably a coding
         bug anyways. *)
      if proveLe e1 e2 why then
        []
      else
        [c]

  | CNullOrLeqNT (z, _, _, _, _) when isZero z -> []

  | CNullOrLeqNT (_, e1, e2, _, why)
  | CLeqNT (e1, e2, _, why) ->
      if proveLe ~allowGt:true e1 e2 why then
        []
      else
        [c]
  | CEq (e1, e2) -> 
      if compareExpStripCasts e1 e2 then []
      else begin
        warn "CEq: Couldn't prove %a  ==  %a. Inserting a runtime check.\n"
          d_plainexp e1 d_plainexp e2;
        [c]
      end
  | CLeqInt (e1, e2, why) -> begin
      match isInteger (constFold true e1), isInteger (constFold true e2) with
        Some n1, Some n2 -> begin
          let n1 = to_int n1 in 
          let n2 = to_int n2 in
          assert (n2 >= 0);
          if n1 < 0 || n1 > n2 then 
            error "Bounds check (%s) will always fail: (unsigned)%a <= %a"
              why dx_exp e1 dx_exp e2;
          []
        end
      | _ -> [c]
  end
  | CLessInt (e1, e2, why) -> begin
      match isInteger (constFold true e1), isInteger (constFold true e2) with
        Some n1, Some n2 -> begin
          let n1 = to_int n1 in 
          let n2 = to_int n2 in
          assert (n2 >= 0);
          if n1 < 0 || n1 >= n2 then 
            error "Bounds check (%s) will always fail: (unsigned)%a < %a"
              why dx_exp e1 dx_exp e2;
          []
        end
      | _ -> [c]
  end

  | CPtrArith (lo, hi, p, e, sz) ->
      begin
        match isInteger (constFold true e) with
        | Some n when (to_signedint n) >= 0 ->
            let pBase,  pOff  = getBaseOffset p in
            let hiBase, hiOff = getBaseOffset hi in
            let intFold e = isInteger (constFold true e) in
            if compareExpStripCasts pBase hiBase then
              match intFold e with
              | Some e' ->
                  if (to_signedint e') <= (hiOff - pOff) then
                    []
                  else begin
                    warn "BAD: Bounds check will always fail: %a <= (%a - %a)"
                          dx_exp e dx_exp hi dx_exp p;
                    [c]
                  end
              | None ->
                  optimizeCheck 
                    (CLeqInt (e, integer (hiOff - pOff), "upper bound check"))
            else
              let diff = BinOp (MinusPP, hi, p, uintType) in
              optimizeCheck (CLeqInt (e, diff, "upper bound check"))
        | _ -> [c]
      end

  | CWriteNT (_, _, z, _) when isZero z -> []

  | _ -> [c]

let opt_deputy_instr i =
  match i with
    Call(Some(lv), Lval(Var vf,NoOffset), el, l) ->
      if not(vf.vname = "deputy_max") then [i] else
	(match el with [e1;e2] ->
	  if compareExpStripCasts e1 e2
	  then [Set(lv,e1,l)]
	  else if proveLe ~allowGt:true e1 e2 "Deputy_max"
	  then [Set(lv,e2,l)]
	  else if proveLe ~allowGt:true e2 e1 "Deputy_max"
	  then [Set(lv,e1,l)]
	  else (if !debug then ignore(E.log "opt_deputy_instr: must keep %a\n" d_instr i);
		[i])
	| _ -> (if !debug then ignore(E.log "opt_deputy_instr: not right form %a\n" d_instr i);
		[i]))
  | _ -> (if !debug then ignore(E.log "opt_deputy_instr: not deputy_max: %a\n" d_instr i);
	  [i])

let optimizeVisitor = object (self)
  inherit nopCilVisitor

  method vinst i =
    match instrToCheck i with
    | Some c -> ChangeTo (List.map checkToInstr (optimizeCheck c))
    | None -> if is_deputy_instr i 
    then ChangeTo(opt_deputy_instr i)
    else DoChildren

  method vfunc fd =
    curFunc := fd;
    let cleanup x =
      curFunc := dummyFunDec;
      x
    in
    ChangeDoChildrenPost (fd, cleanup)

end

let false_cond (c:check) = false

let can_elim_check c =
  match optimizeCheck c with
    [] -> true
  | _ -> false

(* map f to all the expressions in a check 
 * only if cnd c is false *)
(* (check -> bool) -> (exp -> exp) -> check -> check *)
let map_to_check ?cond:(cnd=false_cond) f c =
  match c with
    CNonNull e ->
      let e' = f e in
      CNonNull e'
  | CEq(e1,e2) ->
      let e1' = f e1 in
      if cnd (CEq(e1',e2)) then CEq(e1',e2) else
      let e2' = f e2 in
      if cnd (CEq(e1,e2')) then CEq(e1,e2') else
      CEq(e1',e2')
  | CMult(e1,e2) ->
      let e1' = f e1 in
      let e2' = f e2 in
      CMult(e1',e2')
  | CPtrArith(e1,e2,e3,e4,sz) ->
      let e1' = f e1 in
      if isZero e1' then CPtrArith(e1',e2,e3,e4,sz) else
      let e2' = f e2 in
      if cnd (CPtrArith(e1',e2',e3,e4,sz)) then CPtrArith(e1',e2',e3,e4,sz) else
      let e3' = f e3 in
      if cnd (CPtrArith(e1',e2,e3',e4,sz)) then CPtrArith(e1',e2,e3',e4,sz) else
      let e4' = f e4 in
      if cnd (CPtrArith(e1',e2,e3',e4',sz)) then CPtrArith(e1',e2,e3',e4',sz) else
      CPtrArith(e1',e2',e3',e4',sz)
  | CPtrArithNT(e1,e2,e3,e4,sz) ->
      let e1' = f e1 in
      if isZero e1' then CPtrArithNT(e1',e2,e3,e4,sz) else
      let e2' = f e2 in
      if cnd (CPtrArithNT(e1',e2',e3,e4,sz)) then CPtrArithNT(e1',e2',e3,e4,sz) else
      let e3' = f e3 in
      if cnd (CPtrArithNT(e1',e2,e3',e4,sz)) then CPtrArithNT(e1',e2,e3',e4,sz) else
      let e4' = f e4 in
      if cnd (CPtrArithNT(e1',e2,e3',e4',sz)) then CPtrArithNT(e1',e2,e3',e4',sz) else
      CPtrArithNT(e1',e2',e3',e4',sz)
  | CLessInt(e1,e2,why) ->
      let e1' = f e1 in
      let e2' = f e2 in
      CLessInt(e1',e2',why)
  | CLeqInt(e1,e2,why) ->
      let e1' = f e1 in
      let e2' = f e2 in
      CLeqInt(e1',e2',why)
  | CLeq(e1,e2,why) ->
      let e1' = f e1 in
      if cnd (CLeq(e1',e2,why)) then CLeq(e1',e2,why) else
      let e2' = f e2 in
      if cnd (CLeq(e1,e2,why)) then CLeq(e1,e2',why) else
      CLeq(e1',e2',why)
  | CLeqNT(e1,e2,sz,why) ->
      let e1' = f e1 in
      if cnd (CLeqNT(e1',e2,sz,why)) then CLeqNT(e1',e2,sz,why) else
      let e2' = f e2 in
      if cnd (CLeqNT(e1,e2',sz,why)) then CLeqNT(e1,e2',sz,why) else
      CLeqNT(e1',e2',sz,why)
  | CNullOrLeq(e1,e2,e3,why) ->
      let e1' = f e1 in
      if isZero e1' then CNullOrLeq(e1',e2,e3,why) else
      let e2' = f e2 in
      if cnd (CNullOrLeq(e1',e2',e3,why)) then CNullOrLeq(e1',e2',e3,why) else
      let e3' = f e3 in
      if cnd (CNullOrLeq(e1',e2,e3',why)) then CNullOrLeq(e1',e2,e3',why) else
      CNullOrLeq(e1',e2',e3',why)
  | CNullOrLeqNT(e1,e2,e3,sz,why) ->
      let e1' = f e1 in
      if isZero e1' then CNullOrLeqNT(e1',e2,e3,sz,why) else
      let e2' = f e2 in
      if cnd (CNullOrLeqNT(e1',e2',e3,sz,why)) then CNullOrLeqNT(e1',e2',e3,sz,why) else
      let e3' = f e3 in
      if cnd (CNullOrLeqNT(e1',e2,e3',sz,why)) then CNullOrLeqNT(e1',e2,e3',sz,why) else
      CNullOrLeqNT(e1',e2',e3',sz,why)
  | CWriteNT(e1,e2,e3,sz) ->
      let e1' = f e1 in
      let e2' = f e2 in
      let e3' = f e3 in
      CWriteNT(e1',e2',e3',sz)
  | CSelected e ->
      let e' = f e in
      CSelected e'
  | CNotSelected e ->
      let e' = f e in
      CNotSelected e'
  | CNullUnionOrSelected (lv,e) ->
     let e' = f e in
     CNullUnionOrSelected(lv,e')

(* Apply f to all expressions in a check
 * and return the combination of the results
 * specified by comb the default for which is
 * ||.
 *)
(* (bool -> bool -> bool) -> (exp -> bool) -> check -> bool *)
let test_check ?comb:(cmb=(fun a b -> a || b)) f c =
  match c with
    CNonNull e
  | CNotSelected e
  | CSelected e 
  | CNullUnionOrSelected (_,e) ->
      f e
  | CEq(e1,e2)
  | CMult(e1,e2)
  | CLessInt(e1,e2,_)
  | CLeq(e1,e2,_)
  | CLeqInt(e1,e2,_)
  | CLeqNT(e1,e2,_,_) ->
      let b1 = f e1 in
      let b2 = f e2 in
      cmb b1 b2
  | CNullOrLeq(e1,e2,e3,why) ->
      let b1 = f e1 in
      let b2 = f e2 in
      let b3 = f e3 in
      cmb b1 (cmb b2 b3)
  | CNullOrLeqNT(e1,e2,e3,sz,_)
  | CWriteNT(e1,e2,e3,sz) ->
      let b1 = f e1 in
      let b2 = f e2 in
      let b3 = f e3 in
      cmb (cmb b1 b2) b3
  | CPtrArith(e1,e2,e3,e4,sz)
  | CPtrArithNT(e1,e2,e3,e4,sz) ->
      let b1 = f e1 in
      let b2 = f e2 in
      let b3 = f e3 in
      let b4 = f e4 in
      cmb (cmb b1 b2) (cmb b3 b4)

let checkVisit_change = ref true

let checkAEVisit action (fd : fundec) = object(self)
  inherit AE.aeVisitorClass as super

  method private do_action b e =
    match self#get_cur_eh() with
      None -> e
    | Some eh -> 
	let e', b' = action eh sid e fd b in
	if b' then checkVisit_change := true;
	e'

  method private fix_check =
    let fold_act e = constFold true (self#do_action true e) in
    map_to_check ~cond:can_elim_check fold_act

  method vexpr e = 
    ChangeTo (self#do_action false e)

  method vinst i =
    let action = super#vinst i in
    if action <> DoChildren then
      E.s (bug "Expected DoChildren");
    match instrToCheck i with
    | Some c ->
        let c' = S.time "handle checks" self#fix_check c in
        ChangeTo [checkToInstr c']
    | None ->
        DoChildren

  method vfunc fd =
    S.time "available expressions" AE.computeAEs fd;
    curFunc := fd;
    let cleanup x =
      curFunc := dummyFunDec;
      x
    in
    ChangeDoChildrenPost (fd, cleanup)

end

(* applies the action to the function until
   no changes are made, or lim is reached *)
(* action: 'a -> sid -> exp -> fundec -> bool -> exp * bool *)
(* action -> int -> fundec -> unit *)
let fp cls action lim fd =
  let vis = cls action fd in
  let i = ref 0 in
  checkVisit_change := true;
  while !i < lim && !checkVisit_change do
    if !debug then ignore(E.log "fp: in while loop\n");
    checkVisit_change := false;
    ignore(visitCilFunction (vis :> cilVisitor) fd);
    i := !i + 1;
  done

(* zra - Try to remove tmp variables in exps
 * through forward substitution.
 *
 * This also does forward subst. of everything
 * into checks, but this should probably be replaced
 * by an accurate constant propagation.
 *
 * Run after optimizeVisitor.
 * Only run if optLevel = 2 
 * Temps might be copied to temps, so it is necessary
 * to reach a fixed point.
 *)
(*let forwardTmpSub fd = 
  let vis = checkVisit RCT.fwd_subst fd in
  ignore(visitCilFunction (vis :> cilVisitor) fd)

let constProp fd =
  let vis = checkVisit RCT.const_prop fd in
  ignore(visitCilFunction (vis :> cilVisitor) fd)*)

let forwardTmpSub = fp checkAEVisit RCT.ae_fwd_subst 4

(* Constant propagation into checks *)
let constProp = fp checkAEVisit RCT.ae_const_prop 4

(**************************************************************************)

(**********************************
 * Flow-sensitive optimizer for
 * removing duplicate checks across
 * multiple statements
 **********************************)

let compareExpLists el1 el2 =
  if not(List.length el1 = List.length el2)
  then false
  else List.fold_left (fun b (e1,e2) ->
    b && compareExpStripCasts e1 e2)
      true (List.combine el1 el2)

let deputyCallsEqual i1 i2 =
  match instrToCheck i1, instrToCheck i2 with
    Some c1, Some c2 -> checks_equal c1 c2
  | Some _, None -> false
  | None, Some _ -> false
  | None, None ->
      if not(is_deputy_instr i1) ||
	not(is_deputy_instr i2)
      then false
      else match i1, i2 with
	Call(_,fn1,el1,_), Call(_,fn2,el2,_) ->
	  if not(compareExp fn1 fn2)
	  then false
	  else compareExpLists el1 el2
      | _ -> false


(* do two lists contain the same checks *)
let il_equals il1 il2 =
  if not(List.length il1 = List.length il2)
  then false
  else List.fold_left (fun b i1 ->
    b && List.exists (fun i2 ->
      deputyCallsEqual i1 i2) 
      il2) 
      true il1

(* return the intersection of two lists
 * of checks *)
let il_combine il1 il2 =
  List.filter (fun i1 ->
    List.exists (fun i2 ->
      deputyCallsEqual i1 i2) il2) il1

(* add new checks from chks to cl *)
(* instr list -> instr list -> instr list *)
let il_add il newil =
  List.fold_left (fun il' i ->
    if not(List.exists (fun i' ->
      deputyCallsEqual i i') il')
    then i::il' else il')
    il newil

let il_pretty () il = 
  line ++ seq line (fun i ->
    (d_instr () i)) il

(* see if f returns true on an
 * expression in the list *)
let expListTest f el =
  List.fold_left (fun b e ->
    b || (f e))
    false el

(* if f is true on an instruction
 * then filter it out of the list *)
let ilKiller f il =
  List.filter (fun i ->
    match instrToCheck i with
      Some c ->
	not(test_check f c)
    | None -> match i with
	Call(_,_,el,_) ->
	  not(expListTest f el)
      | _ -> true)
    il

(* filter out checks having memory reads *)
let il_kill_mem il =
  ilKiller AE.exp_has_mem_read il

(* filter out checks refering to vi *)
let il_kill_vi il vi =
  ilKiller (AE.exp_has_vi vi) il

(* filter out checks refering to lv *)
let il_kill_lval il lv =
  ilKiller (AE.exp_has_lval lv) il

let il_handle_inst i il = 
  if is_check_instr i then il else
  match i with
    Set((Mem _, _),_,_) ->
      il_kill_mem il
  | Set((Var vi, NoOffset),e,_) ->
      (match e with
	Lval(Var vi', NoOffset) ->
	  if vi'.vid = vi.vid then il else
	  il_kill_vi il vi
      | _ -> il_kill_vi il vi)
  | Set(lv,_,_) -> il_kill_lval il lv
  | Call(Some(Var vi, NoOffset),_,_,_) ->
      let il' = il_kill_vi il vi in
      if is_deputy_instr i then il' else
      il_kill_mem il'
  | Call(_,_,_,_) ->
      il_kill_mem il
  | Asm(_,_,_,_,_,_) ->
      let _, d = UD.computeUseDefInstr i in
      UD.VS.fold (fun vi il' ->
	il_kill_vi il' vi) d il

module AvailChecks =
  struct

    let name = "Available Checks"

    let debug = debug

    (* list of checks that are available *)
    type t = instr list

    let copy il = il

    let stmtStartData = IH.create 64

    let pretty = il_pretty

    let computeFirstPredecessor stm il = il

    let combinePredecessors (stm:stmt) ~(old:t) (il:t) =
      if il_equals old il then None else
      Some(il_combine old il)

    let doInstr i il =
      let action il =
        match instrToCheck i with
        | Some _ -> il_add il [i]
        | None -> 
	    if is_deputy_instr i
	    then il_add il [i]
	    else il_handle_inst i il
      in
      DF.Post action

    let doStmt stm il = DF.SDefault

    let doGuard c il = DF.GDefault

    let filterStmt stm = true

  end

module AC = DF.ForwardsDataFlow(AvailChecks)

let computeACs fd =
  try let slst = fd.sbody.bstmts in
  let first_stm = List.hd slst in
  IH.clear AvailChecks.stmtStartData;
  IH.add AvailChecks.stmtStartData first_stm.sid [];
  AC.compute [first_stm]
  with Failure "hd" -> ()
  | Not_found -> ()

let getACs sid =
  try Some(IH.find AvailChecks.stmtStartData sid)
  with Not_found -> None

(* Visitor that eliminates a check at a statement where it is available *)
class dupCheckElimClass = object(self)
  inherit nopCilVisitor

  method private filter_dups ail il =
    (* make in instr that sets the lhs of i
     * to the lhs of ai *)
    let makeSet i ai =
      match i, ai with
	Call(Some lv1,_,_,l),Call(Some lv2,_,_,_) ->
	  Set(lv1,Lval lv2, l)
      | _ -> E.s "dupCheckElim: bad deputy instrs in list\n"
    in
    let rec filter_dups' ail il ril =
      match il with [] -> List.rev ril
      | i::rest -> match instrToCheck i with
	| Some _ ->
	    if List.exists (deputyCallsEqual i) ail
	    then (if !debug then ignore(E.log "dupCheckElim: state=\n%a\n Removing: %a\n"
					  il_pretty ail d_instr i);
		  filter_dups' ail rest ril)
	    else let ail' = il_add ail [i] in
	    (if !debug then ignore(E.log "dupCheckElim: state=\n%a\n Not Removing: %a\n"
				     il_pretty ail d_instr i);
	     filter_dups' ail' rest (i::ril))
	| None -> 
	    if is_deputy_instr i then
	      if List.exists (deputyCallsEqual i) ail
	      then let ai = List.find (deputyCallsEqual i) ail in
	      let newi = makeSet i ai in
	      (if !debug then ignore(E.log "dupCheckElim: state=\n%a\n Removing: %a\n"
				       il_pretty ail d_instr i);
	       filter_dups' ail rest (newi::ril))
	      else let ail' = il_add ail [i] in
	      (if !debug then ignore(E.log "dupCheckElim: state=\n%a\n Not Removing: %a\n"
				       il_pretty ail d_instr i);
	       filter_dups' ail' rest (i::ril))
	    else let ail' = il_handle_inst i ail in
	    filter_dups' ail' rest (i::ril)
    in
    filter_dups' ail il []

  method vstmt s =
    match getACs s.sid with
    | None -> SkipChildren
    | Some acs ->
	match s.skind with
	  Instr il -> 
	    let il' = self#filter_dups acs il in
	    s.skind <- Instr il';
	    SkipChildren
	| _ -> DoChildren

end

let dupCheckElimer = new dupCheckElimClass

let elim_dup_checks fd =
  computeACs fd;
  ignore(visitCilFunction dupCheckElimer fd)

(********************************
 *
 * Move loop invariant checks out
 * of loops.
 *
 ********************************)

(* Is the block reachable through a goto? *)
let hasALabel (b:block) : bool =
  let hasLabel = ref false in
  let hasALabelVisitor = object (self)
    inherit nopCilVisitor
    method vstmt s =
      if s.labels <> [] then begin
        hasLabel := true;
      end;
      DoChildren
  end in
  ignore (visitCilBlock hasALabelVisitor b);
  !hasLabel

(* stm -> UD.VS.t *)
let find_varying loop_stm =
  let _, d = UD.computeDeepUseDefStmtKind loop_stm.skind in
  d

(* for visiting a block to determine whether it contains
 * a write to memory or a function call *)
let has_write_or_call = ref false
class writeOrCallFinderClass = object(self)
  inherit nopCilVisitor
      method vinst i =
	if is_deputy_instr i then SkipChildren else
	match i with
	  Set((Mem _,_),_,_) -> 
	    has_write_or_call := true;
	    DoChildren
	| Call(_,_,_,_) ->
	    has_write_or_call := true;
	    DoChildren
	| Asm(_,_,_,_,_,_) -> (* Not precisely... *)
	    has_write_or_call := true;
	    DoChildren
	| _ -> DoChildren
end

let writeOrCallFinder = new writeOrCallFinderClass

(* block -> bool *)
let block_has_write_or_call b =
  has_write_or_call := false;
  ignore(visitCilBlock writeOrCallFinder b);
  !has_write_or_call

(* UD.VS.t -> fundec -> UD.VS.t *)
let add_globs_addrs vvs fd =
  List.fold_left (fun vvs' vi ->
    if vi.vglob || vi.vaddrof
    then UD.VS.add vi vvs'
    else vvs') vvs (fd.sformals@fd.slocals)

(* check -> bool *)
let check_has_mem_read c =
  test_check AE.exp_has_mem_read c

(* the argument b is the body of a Loop *)
(* returns the loop termination condition along with
 * the checks that must preceed its evaluation *)
(* block -> (exp option * instr list) *)
let get_loop_condition b =
  let isListOfChecks =
    List.fold_left (fun b i ->
      b && (is_check_instr i))
      true
  in
  (* returns the first non-empty, non-check
   * statement of a statement list along with
   * the list of checks that were seen *)
  (* stm list -> stm list * instr list *)
  let rec skipEmpty = function
      [] -> [], []
    | {skind = Instr []; labels = []}::rest ->
	skipEmpty rest
    | ({skind = Instr il; labels = []} as s)::rest ->
	if isListOfChecks il
	then let sl,il' = skipEmpty rest in
	sl, il@il'
	else let sl, _ = skipEmpty rest in
	s::sl, []
    | x -> x, []
  in
  (* stm -> exp option * instr list *)
  let rec get_cond_from_if if_stm =
    match if_stm.skind with
      If(e,tb,fb,_) ->
	let e = stripNopCasts e in
	RCT.fold_blocks tb;
	RCT.fold_blocks fb;
	let tsl, tcil = skipEmpty tb.bstmts in
	let fsl, fcil = skipEmpty fb.bstmts in
	(match tsl, fsl with
	  {skind = Break _} :: _, [] -> Some e, tcil
	| [], {skind = Break _} :: _ -> 
	    Some(UnOp(LNot, e, intType)), fcil
	| ({skind = If(_,_,_,_)} as s) :: _, [] ->
	    let teo, tcil' = get_cond_from_if s in
	    (match teo with
	      None -> None, []
	    | Some te -> 
		Some(BinOp(LAnd,e,stripNopCasts te,intType)), tcil@tcil')
	| [], ({skind = If(_,_,_,_)} as s) :: _ ->
	    let feo, fcil' = get_cond_from_if s in
	    (match feo with
	      None -> None, []
	    | Some fe -> 
		Some(BinOp(LAnd,UnOp(LNot,e,intType),
			   stripNopCasts fe,intType)), fcil@fcil')
	| {skind = Break _} :: _, ({skind = If(_,_,_,_)} as s):: _ ->
	    let feo,fcil' = get_cond_from_if s in
	    (match feo with
	      None -> None, []
	    | Some fe -> 
		Some(BinOp(LOr,e,stripNopCasts fe,intType)), fcil@fcil')
	| ({skind = If(_,_,_,_)} as s) :: _, {skind = Break _} :: _ ->
	    let teo, tcil' = get_cond_from_if s in
	    (match teo with
	      None -> None, []
	    | Some te -> 
		Some(BinOp(LOr,UnOp(LNot,e,intType),
			   stripNopCasts te,intType)), tcil@tcil')
	| ({skind = If(_,_,_,_)} as ts) :: _ , ({skind = If(_,_,_,_)} as fs) :: _ ->
	    let teo, tcil' = get_cond_from_if ts in
	    let feo, fcil' = get_cond_from_if fs in
	    (match teo, feo with
	      Some te, Some fe ->
		Some(BinOp(LOr,BinOp(LAnd,e,stripNopCasts te,intType),
			   BinOp(LAnd,UnOp(LNot,e,intType),
				 stripNopCasts fe,intType),intType)),
		tcil@fcil@tcil'@fcil'
	    | _,_ -> None, [])
	| _, _ -> (if !debug then ignore(E.log "checkMover: branches of %a not good\n"
					   d_stmt if_stm);
		   None, []))
    | _ -> (if !debug then ignore(E.log "checkMover: %a not an if\n" d_stmt if_stm);
	    None, [])
  in
  let sl, cil = skipEmpty b.bstmts in
  match sl with
    ({skind = If(_,_,_,_)} as s) :: _ ->
      let eo, cil' = get_cond_from_if s in
      eo, cil@cil'
  | s :: _ -> 
      (if !debug then ignore(E.log "checkMover: %a is first, not an if\n"
			       d_stmt s);
       None, [])
  | [] ->
      (if !debug then ignore(E.log "checkMover: no statements in loop block?\n");
       None, [])

(* UD.VS.t -> check -> bool *)
let check_has_varying_var vvs c =
  UD.VS.fold (fun vi b ->
    b || test_check (AE.exp_has_vi vi) c)
    vvs false

(* if the instruction is deputy_findnull or
 * deputy_max, then if the arguments
 * are loop invariant and there is no
 * other definition of vi reaching i, then
 * the instruction can be lifted out of the
 * loop *)
let can_lift_instr i vvs mw iosh =
  let bad_args el =
    List.fold_left (fun b e ->
      b || (UD.VS.fold (fun vi b ->
	b || (AE.exp_has_vi vi e) ||
	(mw && AE.exp_has_mem_read e)) vvs false)) 
      false el
  in
  let check_def vi =
    if IH.mem iosh vi.vid then
      let ios = IH.find iosh vi.vid in
      if not(RD.IOS.cardinal ios = 2) then 
	(if !debug then ignore(E.log "checkMover: %d defs reach\n" (RD.IOS.cardinal ios));
	 false)
      else match RD.IOS.elements ios with
	[io1;io2] -> (match io1,io2 with
	  Some did, None |
	  None, Some did -> RD.isDefInstr i did
	| Some d1, Some d2 -> 
	    (if !debug then ignore(E.log "checkMover: more than one real def reaches %d %d\n" d1 d2);
	     false)
	| _ -> false)
      | _ -> (if !debug then ignore(E.log "checkMover: impossible\n");
	      false)
    else (if !debug then ignore(E.log "checkMover: %s not in iosh\n" vi.vname);
	  false)
  in
  match i with
    Call(Some(Var vi,NoOffset),Lval(Var vf,NoOffset),el,_) ->
     if (vf.vname = "deputy_findnull" || vf.vname = "deputy_max")
     then if bad_args el 
     then (if !debug then ignore(E.log "checkMover: has varying arg %a\n" d_instr i);
	   false)
     else if check_def vi
     then (if !debug then ignore(E.log "checkMover: can be moved %a\n" d_instr i);
	   true)
     else (if !debug then ignore(E.log "checkMover: had bad def %a\n" d_instr i);
	   false)
     else false
  | _ -> false	 

(* remove and return a list of check instructions
 * not containing things in vvs and not having
 * memory reads when mw is true.
 *)
(* UD.VS.t -> bool -> (instr * rddat) list -> instr list * instr list *)
let filter_instr_list vvs mw ildl =
  let gir, rir = List.fold_left 
      (fun (gi,ri) (i,(u,s,iosh)) -> match instrToCheck i with
	None -> if can_lift_instr i vvs mw iosh
	then (gi,i::ri)
	else (i::gi,ri)
      | Some c ->
	  (if check_has_varying_var vvs c
	   then (if !debug then ignore(E.log "checkMover: has varying var %a\n" d_instr i);
		 (i::gi,ri))
	   else if mw && check_has_mem_read c
	   then (if !debug then ignore(E.log "checkMover: has memory read %a\n" d_instr i);
		 (i::gi,ri))
	   else (if !debug then ignore(E.log "checkMover: can be moved %a\n" d_instr i); 
		 (gi, i::ri))))
      ([],[])
      ildl
  in
  List.rev gir, List.rev rir

(* remove loop invariant checks from the block,
 * return a list of checks along with the
 * expression that must guard them *)
(* block -> UD.VS.t -> int -> (exp,instr list) list *)
let rec filter_and_collect_checks b vvs sid =
  let mw = block_has_write_or_call b in
  List.fold_left (fun cl s ->
    match s.skind with 
      Block b' -> cl@(filter_and_collect_checks b' vvs s.sid)
    | If(e,tb,fb,_) ->
	(* see if e is loop invariant *)
	let aedato = AE.getAEs s.sid in
	(match aedato with None -> cl | Some aedat ->
	  let e',_ = RCT.ae_simp_fwd_subst aedat e true in
	  let e' = stripNopCasts e' in
	  if (UD.VS.exists (fun vi -> AE.exp_has_vi vi e') vvs) ||
	  (mw && AE.exp_has_mem_read e) then cl (* can't move *)
	  else let tgcl = filter_and_collect_checks tb vvs s.sid in
	  let fgcl = filter_and_collect_checks fb vvs s.sid in
	  let tgcl' = List.map (fun (g,cl) ->
	    if compareExpStripCasts g one
	    then (e',cl)
	    else if compareExpStripCasts g e'
	    then (e',cl)
	    else (BinOp(LAnd,e',g,intType),cl)) tgcl in
	  let fgcl' = List.map (fun (g,cl) ->
	    if compareExpStripCasts g one
	    then (UnOp(LNot,e',intType),cl)
	    else (BinOp(LAnd,UnOp(LNot,e',intType),g,intType),cl))
	      fgcl in
	  cl@tgcl'@fgcl')
    | Instr il ->
	(match RD.getRDs s.sid with None -> cl | Some(_,x,iosh) ->
	  let rd_dat_lst = RD.instrRDs il s.sid ((),x,iosh) false in
	  let il_dat_lst = List.combine il rd_dat_lst in
	  let gi, ri = filter_instr_list vvs mw il_dat_lst in
	  if !debug then ignore(E.log "checkMover: after filter to move %d\n"
				  (List.length ri));
	  s.skind <- Instr gi;
	  cl@[(one ,ri)])
    | _ -> cl) [] b.bstmts

(* exp -> exp * instr list -> stmt option *)
let eil_to_if_stmt cond (e,il) =
  match il with 
    [] -> (if !debug then ignore(E.log "checkMover: no checks to make\n"); 
	   None)
  | _ ->
      let ifblk = mkBlock [mkStmt(Instr il)] in
      if compareExpStripCasts e one ||
      compareExpStripCasts e cond
      then (if !debug then ignore(E.log "checkMover: need check %a\n" d_block ifblk); 
	    Some(mkStmt (Block ifblk)))
      else if compareExpStripCasts e zero 
      then (if !debug then ignore(E.log "checkMover: checks never made\n");
	    None)
      else (if !debug then ignore(E.log "checkMover: need checks %a\n" d_block ifblk);
	    Some(mkStmt (If(e,ifblk,mkBlock [],locUnknown))))

(* stmt option list -> stmt list *)
let rec filter_sol sol = 
  match sol with [] -> [] |
  so :: rest -> match so with
    None -> filter_sol rest
  | Some s -> s :: (filter_sol rest)

(* return an If statement that does
 * the invariant checks if the condition.
 * precede by checks in cil.
 * is true *)
(* instr list -> (exp, instr list) list -> exp -> stmt -> stmt option *)
let make_pre_header cil gcl cond loop_stm =
  let if_stmo_lst = List.map (eil_to_if_stmt cond) gcl in
  let if_stm_lst = filter_sol if_stmo_lst in
  match if_stm_lst with [] -> None | _ ->
    let preheader_checks = mkStmt (Instr cil) in
    let preheader_if = mkStmt (If(cond, mkBlock if_stm_lst, 
      mkBlock [], locUnknown)) in
    let preheader_block = mkBlock [preheader_checks;
				   preheader_if] in
    let preheader_stm = mkStmt (Block preheader_block) in
    Some preheader_stm

let check_moved = ref false
class loopInvCheckMoverClass fd = object(self)
  inherit nopCilVisitor

  val mutable cur_block = ref {bstmts = []; battrs = []}

  method vstmt s = match s.skind with
    Loop(b,_,_,_) ->
      if !debug then ignore(E.log "checkMover: looking at loop statement\n%a\n" d_stmt s);
      let vvs = find_varying s in
      let vvs = if block_has_write_or_call b
      then add_globs_addrs vvs fd else vvs in
      RCT.fold_blocks b;
      let condo, cil = get_loop_condition b in
      (match condo with 
	None -> 
	  if !debug then ignore(E.log "checkMover: sid %d: no cond\n" s.sid);
	  DoChildren 
      | Some cond ->
	  let cond = UnOp(LNot,cond,intType) in (* Need the opposite of termination cond *)
	  if !debug then ignore(E.log "checkMover: found condition %a\n" d_exp cond);
	  let gcl = filter_and_collect_checks b vvs s.sid in
	  if !debug then ignore(E.log "checkMover: %d checks to move\n"
				  (List.length gcl));
	  match gcl with [] -> DoChildren | _ ->
	    let pho = make_pre_header cil gcl cond s in
	    match pho with None -> DoChildren | Some ph ->
	      check_moved := true;
	      if !debug then ignore(E.log "checkMover: pre-header=\n%a\n" d_stmt ph);
	      (* put the new statement in the current
		 block's statement list before the loop *)
	      let rec insert_before seen to_insert notseen =
		match notseen with [] -> seen
		| s'::rest -> if s'.sid = s.sid
		then (if !debug then ignore(E.log "checkMover: inserting pre-header\n");
		      seen@[to_insert]@notseen)
		else insert_before (seen@[s']) to_insert rest
	      in
	      (!cur_block).bstmts <- insert_before [] ph (!cur_block).bstmts;
	      if !debug then ignore(E.log "checkMover: new enclosing block=\n%a\n" d_block (!cur_block));
	      DoChildren)
  | _ -> DoChildren

  method vblock b =
    let old_block = cur_block in
    cur_block <- ref b;
    let block_restore b' =
      cur_block <- old_block;
      b
    in
    ChangeDoChildrenPost(b,block_restore)

  method vfunc fd =
    Cfg.clearCFGinfo fd;
    ignore (Cfg.cfgFun fd);
    AE.computeAEs fd;
    RD.computeRDs fd;
    DoChildren

end

let loopInvCheckMover fd =
  check_moved := true;
  while !check_moved do
    check_moved := false;
    ignore(visitCilFunction (new loopInvCheckMoverClass fd) fd)
  done

class checkMoverCleanupClass = object(self)
  inherit nopCilVisitor

  method private is_empty_if s = 
    match s.skind with
      If(_,tb,fb,_) ->
	(match tb.bstmts, fb.bstmts with
	  [],[] -> true
	| _,_ -> false)
    | _ -> false

  method private if_to_block s =
    if self#is_empty_if s 
    then mkStmt(Block(mkBlock []))
    else s

  method vstmt s = match s.skind with
    If(e,tb,fb,l) ->
      let aedato = AE.getAEs s.sid in
      (match aedato with None -> DoChildren
      | Some aedat -> begin
	  let e',_ = RCT.ae_simp_fwd_subst aedat e true in
	  let e' = stripNopCasts(constFold true e') in
	  match isInteger e' with 
	  | Some 0L when not (hasALabel tb) -> (* tb is unreachable *)
              s.skind <- Block fb;
	      DoChildren
          | Some n when not (hasALabel fb)
              && n <> Int64.zero -> (* fb is unreachable *)
	      s.skind <- Block tb;
	      DoChildren
          | _ -> (* Both branches reachable *)
              s.skind <- If(e',tb,fb,l);
	      ChangeDoChildrenPost(s,self#if_to_block)
        end)
  | _ -> DoChildren

  method vblock b =
    RCT.fold_blocks b;
    let new_stmts = List.filter (fun s ->
      match s.skind with 
	Instr [] when s.labels = [] -> false 
      | _ -> true) b.bstmts in
    b.bstmts <- new_stmts;
    DoChildren

  method vfunc fd =
    Cfg.clearCFGinfo fd;
    ignore (Cfg.cfgFun fd);
    AE.computeAEs fd;
    DoChildren
end

let checkMoverCleanup = new checkMoverCleanupClass


(**************************************************************************)
(**************************************************************************)

(* flow-sensitive optimizer for nonnull, strlen, and inequalities*)


type absState = {
  nonNullVars: varinfo list;

  strlenVars: (varinfo * (exp * referenced)) list; 
  (** (v, e, mem, vars) means that v holds the the length of string e.
      For convenience, we also remember whether e involves a memory reference
      and what local vars e contains. *)

  ineqs: (lval * int64 * lval * referenced) list;
  (** (x, c, y, vars) means that x + c <= y.
    Also, if x is a pointer, then x+c < 2^32, so x+c doesn't overflow.
    (We usually get this fact just by knowing that any ptr arith that appears
     in a program has first been checked for overflow with CPtrArith)
    '+' is done on integers, not pointers.
    We also remember the memory/variables that either expression depend on. *)


  canIncrement: (lval * exp * referenced) list;
  (** (x, e) means that x is an NT pointer that can safely be incremented
      by e. *)
}

let top = { nonNullVars = [];
            strlenVars = [];
            ineqs = [];
            canIncrement = [];
          }

let printIneq (lv1,i,lv2,_): doc = 
  dprintf "%a + %Ld <= %a" dx_lval lv1 i dx_lval lv2

let d_state () a: doc =
  if a.strlenVars == [] then
    dprintf 
      "@[Nonnull vars:@[%a@]\nInEqs:@[%a@]\nCanIncr:@[%a@]@]"
      (docList (fun vi -> text vi.vname)) a.nonNullVars
      (docList ~sep:Pretty.line printIneq) a.ineqs
      (docList ~sep:Pretty.line (fun (x,e,_) -> dprintf "%a by %a"
                                   dx_lval x dx_exp e)) a.canIncrement
  else
    dprintf 
      "@[Nonnull vars:@[%a@]\nStrlenVars:@[%a@]\nInEqs:@[%a@]\nCanIncr:@[%a@]@]"
      (docList (fun vi -> text vi.vname)) a.nonNullVars
      (docList (fun (vi,(e,_)) -> 
                  text(vi.vname^" for ") ++ dx_exp () e))  a.strlenVars
      (docList ~sep:Pretty.line printIneq) a.ineqs
      (docList ~sep:Pretty.line (fun (x,e,_) -> dprintf "%a by %a"
                                   dx_lval x dx_exp e)) a.canIncrement
      

(* This fake variable is used to represent 0.  It should never appear in
   output code *)
let zeroLv : lval =
  let v = makeVarinfo false "_ZERO_" intType in
  var v


let addNonNull (a:absState) (vi: varinfo) : absState = 
  if List.memq vi a.nonNullVars then a
  else begin
(*     log "%s is nonnull.\n" vi.vname; *)
    { a with nonNullVars = vi::a.nonNullVars }
  end

let addStringlen (a:absState) (vi:varinfo) (str:exp) : absState =
  try 
    let other,_ = List.assq vi a.strlenVars in
    if compareExpStripCasts other str then
      a
    else
      E.s (unimp "%s is the length of two different strings: %a and %a\n"
             vi.vname dx_exp other dx_exp str)
  with Not_found ->
    let kill : referenced = Dutil.varsOfExp str in          
    { a with strlenVars = (vi,(str,kill))::a.strlenVars }

(* Do we know the length of this string? *)
let hasStringlen (a:absState) (str:exp) : exp option =
  let rec loop = 
    function (vi, (e, _))::rest ->
      if compareExpStripCasts e str then 
        Some (Lval (var vi)) 
      else loop rest
    | [] -> None
  in
  loop a.strlenVars

let addLessEq (a:absState) (lv1: lval) (i:int64) (lv2:lval) : absState = 
  let alreadyExists: bool =
    (* unlike ineqHolds, we look for this exact statement rather than a
       stronger one. *)
    List.exists
      (fun (lv1', i', lv2', _) ->
         (compareLval lv1 lv1') && (compareLval lv2 lv2') && i' = i)
      a.ineqs
  in
  if alreadyExists then a
  else begin
    (* make a fake binop so we can get the combined deps of e1 and e2*)
    let e' = BinOp(PlusA, Lval lv1, Lval lv2, intType) in
    let fact = (lv1, i, lv2, Dutil.varsOfExp e') in
    { a with ineqs = fact::a.ineqs }
  end
    
(* Is it true that (lv1 + i <= lv2)?
 * To answer this, we look for (lv1, i', lv2) in the state, where i' >= i. *)
let ineqHolds (a:absState) (lv1: lval) (i:int64) (lv2:lval) : bool =
  let findIneq lv1 i lv2: bool =
    (* Search the state for lv1 + i <= lv2, or a stronger statement. *)
    List.exists
      (fun (lv1', i', lv2', _) ->
         (compareLval lv1 lv1') && (compareLval lv2 lv2') && i' >= i)
      a.ineqs
  in
(*   log "Checking inequality %a + %Ld <= %a\n in %a" *)
(*     dx_lval lv1 i dx_lval lv2 d_state a; *)
  if compareLval lv1 lv2 then
    (i <=% 0L)
  else if findIneq lv1 i lv2 then
    true
  else if lv1 != zeroLv && lv2 != zeroLv then begin
    (* One last shot: If there exists k such that lv1 <= k &&  k + i <= lv2,
       then the inequality holds. *)
    List.exists
      (fun (lv1', negk, z, _) ->
         z == zeroLv && (compareLval lv1 lv1')
           (* so far, lv + negk <= 0 (i.e. lv <= -negk)
              Now check k + i <= lv2. *)
         && findIneq zeroLv (i -% negk) lv2)
      a.ineqs
  end else
    false


let addCanIncrement (a:absState) (p:lval) (howmuch:exp) : absState =
 (* make a fake binop so we can get the combined deps of p and howmuch *)
  let e' = BinOp(PlusA, Lval p, howmuch, intType) in
  let fact = (p, howmuch, Dutil.varsOfExp e') in
  if List.mem fact a.canIncrement then a
  else 
    { a with canIncrement = fact::a.canIncrement }

let canIncrement (a:absState) (p:lval) (howmuch:exp) : bool =
  (* TODO: is it worthwhile to use inequality information here? *)
  (List.exists
     (fun (p', e', _) ->
        (compareLval p p') && (compareExpStripCasts howmuch e') )
     a.canIncrement)
  || 
    (* Is howmuch == strlen(p)? *)
    ((hasStringlen a (Lval p)) = (Some howmuch))

(* Remove certain entries from a list, without preserving order *)
let removeAll (f: 'a -> bool) (l: 'a list): 'a list =
  if List.exists f l then
    List.fold_left
      (fun acc x -> if f x then acc else x::acc)
      []
      l
  else
    l

let updateNonnull (a:absState) nonnull' : absState =
  if nonnull' == a.nonNullVars then a
  else
    { a with nonNullVars = nonnull' }
let updateStrlen (a:absState) strlens' : absState =
  if strlens' == a.strlenVars then a
  else 
    { a with strlenVars = strlens' }
let updateIneqs (a:absState) ineqs' : absState =
  if ineqs' == a.ineqs then a
  else 
    { a with ineqs = ineqs' }
let updateCanIncr (a:absState) ci' : absState =
  if ci' == a.canIncrement then a
  else 
    { a with canIncrement = ci' }

let scrambleVar (a:absState) (v: varinfo) : absState =
  let a =
    let nonnull' = removeAll ((==) v) a.nonNullVars in
    updateNonnull a nonnull'
  in
  let a = 
    let refersToV (v',(_,kill)) = (v==v') || (List.memq v kill.varsRead) in
    let strlen' = removeAll refersToV a.strlenVars in
    updateStrlen a strlen'
  in
  let a = 
(*     let refersToV ((_,_,_,kill) as ie) =  *)
(*       log "Considering %a, kill %s" insert (printIneq ie) v.vname; *)
(*       let res = (List.memq v kill.varsRead) in *)
(*       if res then log "Killing %a\n" insert (printIneq ie) *)
(*       else log "not killing %a, deps on %a\n" insert (printIneq ie) *)
(*         (docList (fun v -> text v.vname)) kill.varsRead; *)
(*       res *)
(*     in *)
    let refersToV (_,_,_,kill) = (List.memq v kill.varsRead) in
    let ineqs' = removeAll refersToV a.ineqs in
    updateIneqs a ineqs'
  in
  let a = 
    let refersToV (_,_,kill) = (List.memq v kill.varsRead) in
    let ci' = removeAll refersToV a.canIncrement in
    updateCanIncr a ci'
  in
  a  

(* After a write to memory, scramble all vars whose addresses have been
   taken.*)
let scrambleMem ?(globalsToo=false) (a:absState) : absState =
  let scrambled: varinfo -> bool = 
    fun vi -> vi.vaddrof || (vi.vglob && globalsToo)
  in
  let a =
    let nonnull' = removeAll scrambled a.nonNullVars in
    updateNonnull a nonnull'
  in
  let a = 
    let refersToMem (v',(_,kill)) = 
      kill.memRead || (scrambled v') || (List.exists scrambled kill.varsRead) in
    let strlen' = removeAll refersToMem a.strlenVars in
    updateStrlen a strlen'
  in
  let a = 
    let refersToMem (_,_,_,kill) = 
      kill.memRead || (List.exists scrambled kill.varsRead) in
    let ineqs' = removeAll refersToMem a.ineqs in
    updateIneqs a ineqs'
  in
  let a = 
    let refersToMem (_,_,kill) = 
      kill.memRead || (List.exists scrambled kill.varsRead) in
    let ci' = removeAll refersToMem a.canIncrement in
    updateCanIncr a ci'
  in
  a  
  
let stateMap : absState IH.t = IH.create 50

let isNonNull state e: bool = 
(*   log "isNonNull? on %a.\n" d_plainexp e; *)
  (isNonnullType (typeOf e)) ||
  (match stripNopCasts e with
    Lval(Var vi, NoOffset) -> 
      List.memq vi state.nonNullVars 
  | BinOp((PlusPI|IndexPI|MinusPI), e1, e2, _) ->
      (* We've disallowed ptr arith if e1 is null. *)
      let t1 = typeOf e1 in
      isPointerType t1 && not (isSentinelType t1)
  | AddrOf lv
  | StartOf lv -> true
  | Const(CStr _) -> true
  | _ -> false)

let isFalse state e =
  match e with
    UnOp(LNot, e', _) -> isNonNull state e'
  | _ -> isZero e


exception CantDecompose

(* Take this sum expression and return p, [off1; off2; ...], c
    where e = p + (off1 + off2 + ... + c), p is a pointer or zeroLv, 
    and the first + is ptr arith (unless p is zeroLv). *)
let rec expToSum (e: exp) : lval * (lval list) * int64 =
  let e' = constFold true (stripCastsForPtrArith e) in
  match e' with
    Lval lv
  | StartOf lv -> (* Treat array references as lvals. *)
      (* we get the type of e, not e', because we want to consider casts
         when deciding whether to classify this as a pointer or an int. *)
      if isPointerType (typeOf e) then
        lv, [], 0L
      else if isArrayType (typeOf e) then
        E.s (bug "expToSum on %a" dx_exp e)
      else
        zeroLv, [lv], 0L
  | Const _ when isIntegralType (typeOf e) ->
      let c = match isInteger e' with
          Some i64 -> i64
        | None -> E.s (bug "expected a constant here.")
      in
      zeroLv, [], c
  | BinOp((PlusPI|IndexPI), e1, e2, t) ->
      let lvp1, offs1, c1 = expToSum e1 in
      let lvp2, offs2, c2 = expToSum e2 in
      if lvp2 != zeroLv then
        E.s (bug ("expToSum on \"%a\" returned a pointer, but there should be"
                  ^^" no pointer on the right operand of PlusPI") dx_exp e);
      if lvp1 == zeroLv then
        E.s (unimp "expToSum returned NULL for the left operand of PlusPI");
      lvp1, offs1@offs2, (c1+%c2)
  | BinOp(PlusA, e1, e2, t) when isIntegralType t ->
      let lvp1, offs1, c1 = expToSum e1 in
      let lvp2, offs2, c2 = expToSum e2 in
      if lvp1 != zeroLv || lvp2 != zeroLv then
        E.s (bug ("expToSum on \"%a\" returned a pointer, but there should be"
                  ^^" no pointer with PlusA") dx_exp e);
      zeroLv, offs1@offs2, (c1+%c2)
  | _ ->
      (*if !debug then
        log "Can't decompose %a into a sum." dx_exp e;*)
      raise CantDecompose

(* Given a list of lvals, return an lval representing the sum of the list
   elements, or raise CantDecompose if there are two or more lvals in the list*)
let getOneLval (l:lval list) : lval =
  match l with
    [] -> zeroLv
  | [lv] -> lv
  | _ -> raise CantDecompose


(* Decompose the expression into lv+c (integer arithmetic, not pointer), 
   or raise an exception if this isn't possible *)
let rec expToIntSum (e:exp) : lval * int64 = 
  match expToSum e with
    lv, [], c when lv != zeroLv -> 
      (* FIXME: I'm trusting here that there's no overflow, since 
         that should already be taken care of.  Is there a better way to reason
         about overlfow? *)
      (* This is lv+c.  Do the multiplication now, to handle ptr arith. *)
      if c <> 0L then
        lv, (c *% (Int64.of_int (baseSize (typeOfLval lv))))
      else
        (* if c = 0, don't compute the baseSize, since we don't need to
           (and it could be a pointer to an abstract type.) *)
        lv, 0L
  | z, offs, c when z == zeroLv ->
      (* offs + c, no pointers involved. *)
      getOneLval offs, c
  | _ -> 
      raise CantDecompose



(* FIXME: use the sign info!  E.g. add e1 >= 0 *)
let doLessEq a (e1: exp) (e2:exp) ~(signed:bool): absState = 
(*   log "Guard %a <= %a.\n" dx_exp e1 dx_exp e2; *)
  try 
    let lv1, c1 = expToIntSum e1 in
    let lv2, c2 = expToIntSum e2 in
    addLessEq a lv1 (c1 -% c2) lv2
  with CantDecompose -> a


let doLess a (e1: exp) (e2:exp) ~(signed:bool): absState = 
(*   log "Guard %a < %a.\n" d_plainexp e1 d_plainexp e2; *)
  try 
    let lv1, c1 = expToIntSum e1 in
    let lv2, c2 = expToIntSum e2 in
    addLessEq a lv1 (1L +% c1 -% c2) lv2
  with CantDecompose -> a

(* Update a state to reflect a branch *)
let doBranch (a:absState) (e:exp) : absState = 
  if !debug then
    log "Guard %a.\n" dx_exp e;
  let e = match stripNopCasts e with
      UnOp(LNot, UnOp(LNot, e, _), _) -> e
    | BinOp(Ne, e, z, _) when isZero z -> e
    | UnOp(LNot, BinOp(Lt, e1, e2, t), _) ->
        BinOp(Ge, e1, e2, t)
    | UnOp(LNot, BinOp(Le, e1, e2, t), _) ->
        BinOp(Gt, e1, e2, t)
    | UnOp(LNot, BinOp(Gt, e1, e2, t), _) ->
        BinOp(Le, e1, e2, t)
    | UnOp(LNot, BinOp(Ge, e1, e2, t), _) ->
        BinOp(Lt, e1, e2, t)
    | e -> e
  in
  match e with
    Lval(Var vi, NoOffset) when isPointerType vi.vtype -> 
      addNonNull a vi
  | Lval(Mem (Lval lvp), NoOffset) when isNullterm (typeOfLval lvp) ->
      addCanIncrement a lvp Cil.one
  | BinOp(Lt, e1, e2, t) when isIntOrPtrType (typeOf e1) ->
      doLess a e1 e2 ~signed:(isSignedType (typeOf e1))
  | BinOp(Le, e1, e2, t) when isIntOrPtrType (typeOf e1) ->
      doLessEq a e1 e2 ~signed:(isSignedType (typeOf e1))
  | BinOp(Gt, e1, e2, t) when isIntOrPtrType (typeOf e1) ->
      doLess a e2 e1 ~signed:(isSignedType (typeOf e1))
  | BinOp(Ge, e1, e2, t) when isIntOrPtrType (typeOf e1) ->
      doLessEq a e2 e1 ~signed:(isSignedType (typeOf e1))
  | _ -> 
      a

(* Add that 
 * lv >= e1 and
 * lv >= e2
 *)
let doMax a lv e1 e2 =
  let a' = doLessEq a e1 (Lval lv) ~signed:(isSignedType(typeOf e1)) in
  let a' = doLessEq a' e2 (Lval lv) ~signed:(isSignedType(typeOf e2)) in
  a'


(* Update a state to reflect a check *)
let processCheck a (c:check) : absState =
  match c with
    CNonNull e -> doBranch a e 
  | CLeq(e1, e2, _) -> doLessEq a e1 e2 ~signed:false
  | _ -> a

(* Add to anext any relevant inequality information for the assignment
     dest := e     
*)
let doSet ~(aold:absState) ~(anext:absState) (dest: lval) (e:exp) : absState =
  if !debug then log "doSet: %a := %a\n" dx_lval dest dx_exp e;
  match isInteger (constFold true e) with
    Some i ->
      (* Add dest <= i and i <= dest *)
      let anext = addLessEq anext dest (~-% i) zeroLv in
      let anext = addLessEq anext zeroLv i dest in
      anext
  | None -> begin
      let anext = match e with
          Lval lv ->
            (* For each fact about lv in the old state,
               add an analogous fact about dest *)
            List.fold_left 
              (fun anext (lv1, e, _) -> 
                 if compareLval lv1 lv then
                   addCanIncrement anext dest e
                 else
                   anext)
              anext
              aold.canIncrement
        | _ -> anext
      in
      try 
        let lvrhs, c = expToIntSum e in
        if c = 0L then begin
          (* The assignment is dest := lvrhs *)

          (* For each fact about lvrhs in the old state,
             add an analogous fact about dest *)
          List.fold_left 
            (fun anext (lv1, n, lv2, _) -> 
               let anext = if compareLval lv1 lvrhs then
                 addLessEq anext dest n lv2
               else anext in
               let anext = if compareLval lv2 lvrhs then
                 addLessEq anext lv1 n dest
               else anext in
               anext)
            anext
            aold.ineqs
        end
        else begin
          (* The assignment is dest := lvrhs+c, where c is an int *)

          (* For each fact about lvrhs in the old state,
             add a fact about dest, after adjusting by c *)
          List.fold_left 
            (fun anext (z, n, lv2, _) -> 
               (* We have to be very conservative, because of
                  overflow.  If 0+n <= lvrhs AND lvrhs <= MAXINT-c,
                  then "dest := lvrhs+c" does not overflow, and we
                  get the new fact 0 + (n+c) <= dest *)
               if (z == zeroLv && compareLval lv2 lvrhs)
                 (* so far, 0+n <= lvrhs *)
                 && 
                 (ineqHolds aold lvrhs (c -% (maxUnsignedInt(typeOf e))) zeroLv)
                 (* now lvrhs <= MAXINT-c.  To see this, note that the line
                    above tests lvrhs + c-MAXINT <= 0, i.e. lvrhs <= -(c-MAXINT)
                 *)
               then
                 addLessEq anext zeroLv (n +% c) dest
               else
                 anext)
            anext
            aold.ineqs
        end
      with CantDecompose -> 
        anext
    end


let flowHandleInstr a i =
  (* E.log "Doing instr %a in state %a\n" d_instr i d_state a; *)
  match instrToCheck i with
  | Some c -> processCheck a c
  | None ->
      begin
        match i with
          Set ((Var vi, NoOffset) as dest, e, _) 
          when isPointerType vi.vtype ->
            let anext = scrambleVar a vi in
            let anext = if isNonNull a e then
              addNonNull anext vi
            else
              anext
            in
            doSet ~aold:a ~anext dest e
        | Set ((Var vi, _) as dest, e, _)-> 
            let anext = scrambleVar a vi in
            doSet ~aold:a ~anext dest e
        | Set ((Mem _, _), _, _)-> 
            scrambleMem a
        | Call (Some(Var vi, NoOffset), Lval (Var vf, NoOffset),
                [estr; bytes], _)
            when (vf.vname = "deputy_findnull") -> begin
              if not (isIntegralType vi.vtype) || not (isPointer estr) then
                E.s (bug "bad arg to %s\n" vf.vname);
              let a = scrambleVar a vi in
              addStringlen a vi estr
            end
        | Call (Some(Var vi, NoOffset), Lval (Var vf, NoOffset),
                [estr], _)
            when (vf.vname = "strlen") -> begin
              if not (isIntegralType vi.vtype) 
                || ((baseSize (typeOf estr)) <> 1) then
                  E.s (bug "bad arg to %s\n" vf.vname);
              let a = scrambleVar a vi in
              addStringlen a vi estr
            end
	| Call (Some(Var vi, NoOffset), Lval(Var vf, NoOffset),
		[e1;e2], _) when vf.vname = "deputy_max" -> begin
		  let a' = scrambleVar a vi in
		  doMax a' (Var vi, NoOffset) e1 e2
		end
        | Call (Some (Var vi, NoOffset), f, _, _)
          when isPointerType vi.vtype ->
	    let a = 
              if is_deputy_instr i then scrambleVar a vi else
                scrambleMem ~globalsToo:true (scrambleVar a vi)
            in
            let rt, _, _, _ = splitFunctionType (typeOf f) in
            if isNonnullType rt then
              addNonNull a vi
            else
              a
        | Call (_, _, _, _) ->
	    scrambleMem ~globalsToo:true a
        | Asm (_, _, writes, _, _, _) -> 
            (* This is a quasi-sound handling of inline assembly *)
            let a = scrambleMem a in
            List.fold_left (fun a (_,_,(lh,_)) -> 
              match lh with Var vi -> scrambleVar a vi
              | Mem _ -> a)
              a
              writes
      end


module Flow = struct
  let name = "DeputyOpt"
  let debug = debug
  type t = absState
  let copy x = x
  let stmtStartData = stateMap
  let pretty = d_state
  let computeFirstPredecessor s a = a

  let combinePredecessors s ~(old:t) newa = 
    let nnv = List.filter 
                (fun vi -> List.memq vi newa.nonNullVars) 
                old.nonNullVars
    in
    let sv = List.filter 
                (fun x -> List.mem x newa.strlenVars) 
                old.strlenVars
    in
    let ie = List.filter 
                (fun ((lv1, c, lv2, _) as x) -> 
                 (* Make sure newa contains something at least as strong as x *)
                   (List.memq x newa.ineqs) (* for performance *)
                   || (ineqHolds newa lv1 c lv2)) 
                old.ineqs
    in
    let ci = List.filter 
                (fun x -> List.mem x newa.canIncrement) 
                old.canIncrement
    in
    if (List.length nnv <> List.length old.nonNullVars) 
      || (List.length sv <> List.length old.strlenVars) 
      || (List.length ie <> List.length old.ineqs)  
      || (List.length ci <> List.length old.canIncrement) then
      Some {nonNullVars = nnv; strlenVars = sv; ineqs = ie; canIncrement = ci}
    else
      None (* at fixed point *)

  let doInstr i a = 
(*     log "Visiting %a  State is %a.\n" dn_instr i d_state a; *)
    DF.Done (flowHandleInstr a i)

  let doStmt s a = 
    curStmt := s.sid;
    DF.SDefault

  let doGuard e a = 
    if isFalse a e then DF.GUnreachable
    else DF.GUse (doBranch a e)

  let filterStmt s = true
end

module FlowEngine = DF.ForwardsDataFlow (Flow)

let flowOptimizeCheck (c: check) ((inState, acc):(absState * check list))
  : (absState * check list) =
  let isNonNull = isNonNull inState in
  (* Returns true if CPtrArith(lo, hi, p, Lval x, sz) can  be optimized away:*)
  let checkPtrArith lo hi p e : bool = 
    match e with
      Lval x ->
       let bp, cur_off = getBaseOffset (stripNopCasts p) in
       let bh, len = getBaseOffset (stripNopCasts hi) in
       let cur_off = Int64.of_int cur_off in
       let len = Int64.of_int len in
       (compareExpStripCasts bp bh)
       && (compareExpStripCasts bp lo)
         (* the check is CPtrArith(lo, lo+len, lo+cur_off, x, sz).
            if -cur_off <= x <= len-cur_off, the check succeeds. *)
       && (ineqHolds inState zeroLv (~-% cur_off) x)   (* 0-cur_off <= x *)
       && (ineqHolds inState x (cur_off -% len) zeroLv)(* x + cur_off-len <= 0,
                                                          so x <= len-cur_off*)
    | _ -> false
  in
  (* Returns true if CLeq(e1, e2) can  be optimized away:*)
  let checkLeq e1 e2 : bool =
    try
      let e1 = stripNopCasts e1 in
      let e2 = stripNopCasts e2 in
      (* log "Checking %a <= %a" dx_exp e1 dx_exp e2; *)
      (* expToSum may raise CantDecompose *)
      let p1, offs1, c1 = expToSum e1 in
      let p2, offs2, c2 = expToSum e2 in
      
      if compareLval p1 p2 then begin
        (* p1 == p2, so we only need to compare offs1+c1 <= offs2+c2 *)
        let off1 = getOneLval offs1 in (* may raise CantDecompose *)
        let off2 = getOneLval offs2 in (* may raise CantDecompose *)
        (*log "%a + %a + %Ld <= %a + %a + %Ld"
          dx_lval p1 dx_lval off1 c1 dx_lval p2 dx_lval off2 c2;*)
        ineqHolds inState off1 (c1 -% c2) off2
      end
      else if offs1 = [] && offs2 = [] then begin
        (* Redo the sum decomposition, this time paying attention to
           pointer arithmetic. *)
        let p1, c1 = expToIntSum e1 in
        let p2, c2 = expToIntSum e2 in
        (* e1 = p1 + c1, and e2 = p2 + c2, using integer arithmetic.
           Now check p1 + c1-c2 <= p2*)
        ineqHolds inState p1 (c1 -% c2) p2
      end else
        false
    with CantDecompose ->
      false
  in

  (* doOpt is called recursivly if we simplify the check to a different check
     that has its own optimization rule. 
     It returns the simplified check, or None if we satisfied the check
     completely.*)
  let rec doOpt (c : check) : check option =
    match c with
    | CNonNull (e1) when isNonNull e1 ->
        None
    | CNonNull (e1) when isZero e1 ->
        error "non-null check will always fail.";
        Some c
    | CNullOrLeq (e1, _, _, why)
    | CNullOrLeqNT (e1, _, _, _, why) when isZero e1 ->
        None
    | CNullOrLeq (e1, e2, e3, why) when isNonNull e1 ->
        doOpt (CLeq(e2, e3, why))
    | CNullOrLeqNT (e1, e2, e3, e4, why) when isNonNull e1 ->
        let c' = CLeqNT(e2, e3, e4, why) in
        doOpt c'
    | CPtrArith(lo, hi, p, e, sz)
    | CPtrArithNT(lo, hi, p, e, sz) when checkPtrArith lo hi p e ->
        None
    | CPtrArithNT(_, _, Lval p, e, _) when canIncrement inState p e ->
        None
    | CLeq(e1, e2, _)
    | CNullOrLeq(_, e1, e2, _)
    | CNullOrLeqNT(_, e1, e2, _, _) when checkLeq e1 e2 ->
        None
    | CLeqNT (e1, e2, sz, why) -> begin
        match hasStringlen inState e2 with
          Some len -> 
            let e2' = BinOp(PlusPI, e2, len, typeOf e2) in
            (* len == strlen(e2).  So rewrite the check as e1 <= e2+len *)
            (* This will often be optimized away completely
               by a later pass, since e1 often equals e2' *)
            doOpt (CLeq(e1, e2', why)) 
        | None -> Some c
      end
    | CLessInt(e1, e2, why) -> begin
        try
          (* expToIntSum may raise CantDecompose *)
          let lv1, c1 = expToIntSum (stripNopCasts e1) in
          let lv2, c2 = expToIntSum (stripNopCasts e2) in
          (* (The "1+" accounts for < vs <=.)  *)
          if (ineqHolds inState lv1 (1L +% c1 -% c2) lv2) (* lv1+c1 < lv2+c2 *)            && (ineqHolds inState zeroLv (~-% c1) lv1)     (* 0 <= lv1+c1 *)
          then
            None
          else
            Some c
        with CantDecompose ->
          Some c
      end
    | CLeqInt(e1, (BinOp (MinusPP, hi, p, uintType)), _) ->
        let e' = BinOp(PlusPI, p, e1, (typeOf p)) in
        if checkLeq e' hi then
          None
        else 
          Some c
    | _ -> Some c
  in
  let acc' = match doOpt c with
     Some c -> c::acc | None -> acc
  in
  (processCheck inState c), acc'


let flowOptimizeVisitor = object (self)
  inherit nopCilVisitor

  val mutable curSid = -1

  method vstmt s =
    (* now that checks and instructions can be mixed,
     * the state has to be changed when an instruction is
     * visited *)
    let rec filterIl state il fl =
      match il with [] -> List.rev fl | i::rest ->
	let new_state = flowHandleInstr state i in
	(match instrToCheck i with
	  Some c ->
	    let _, c' = flowOptimizeCheck c (state,[]) in
	    (match c' with
	      [] -> (if !debug then ignore(E.log "fOV: optimized %a out\n" d_instr i);
		     filterIl new_state rest fl)
	    | [nc] ->
		let i' = checkToInstr nc in
		if !debug then ignore(E.log "fOV: changed to %a out\n" d_instr i');
		filterIl new_state rest (i'::fl)
	    | _ -> (if !debug then ignore(E.log "fOV: didn't optimize %a out\n" d_instr i);
		    filterIl new_state rest (i::fl)))
	| None -> filterIl new_state rest (i::fl))
    in
    begin
      try
        curSid <- s.sid;
        let state = IH.find stateMap s.sid in
        if !debug then
          E.log "Optimizing statement %d with state %a\n" s.sid d_state state;
        begin
          match s.skind with
          | If(e, blk1, blk2, l) when isNonNull state e ->
              if hasALabel blk2 then 
                s.skind <- If(Cil.one, blk1, blk2, l)
              else
                (* blk2 is unreachable *)
                s.skind <- Block blk1
          | If(e, blk1, blk2, l) when isFalse state e ->
              if hasALabel blk1 then
                s.skind <- If(Cil.zero, blk1, blk2, l)
              else
                (* blk1 is unreachable *)
                s.skind <- Block blk2
	  | Instr il ->
	      s.skind <- Instr(filterIl state il [])
          | _ -> ()
        end
      with Not_found -> () (* stmt is unreachable *)
    end;
    DoChildren

  method vfunc fd =
    curFunc := fd;
    let cleanup x = 
      curFunc := dummyFunDec; 
      x
    in
    ChangeDoChildrenPost (fd, cleanup)

end

(** flow-sensitive optimizer for nonnull, strlen, and inequalities*)
let doFlowAnalysis (fd:fundec): unit =
  IH.clear stateMap;
  let fst = List.hd fd.sbody.bstmts in
  IH.add stateMap fst.sid top;
  FlowEngine.compute [fst];
  if !debug then
    E.log "%s: finished analysis; starting optimizations.\n" Flow.name;
  ignore (visitCilFunction flowOptimizeVisitor fd);
  IH.clear stateMap;
  curStmt := -1; 
  ()

(**************************************************************************)

let recomputeCfg (fd:fundec) : unit =
  Cfg.clearCFGinfo fd;
  ignore (Cfg.cfgFun fd)

let optLevel1 (fd:fundec) : unit =
  if !verbose then
    log "Doing flow-insensitive optimizations.\n";
  recomputeCfg fd;
  ignore (visitCilFunction optimizeVisitor fd)

let optLevel2 (fd:fundec) : unit =
  if !verbose then
    log "Doing some flow-sensitive optimizations.\n";
  recomputeCfg fd;
  ignore (visitCilFunction optimizeVisitor fd);
  doFlowAnalysis fd;
  ignore (visitCilFunction optimizeVisitor fd)

let optLevel3 (fd:fundec) : unit =
  if !verbose then
    log "Doing one pass of all optimizations:\n";
  recomputeCfg fd;
  let cf = constFoldVisitor false in
  AE.ignore_inst := is_check_instr;
  AE.ignore_call := is_deputy_instr;
  ignore (S.time "check elimination" (visitCilFunction optimizeVisitor) fd);
  S.time "flow analysis" doFlowAnalysis fd;

  S.time "elim_dup_checks" elim_dup_checks fd;
  S.time "forward substitution" forwardTmpSub fd;
  (*S.time "constant propagation" constProp fd;*)
  ignore(S.time "constant folding" (visitCilFunction cf) fd);
  S.time "elim_dup_checks" elim_dup_checks fd;

  (* doFlowAnalysis *must* come before the optimizeVisitor
     after fwd subst, const prop, and const folding *)
  S.time "flow analysis" doFlowAnalysis fd;
  ignore(S.time "check elimination" (visitCilFunction optimizeVisitor) fd);

  ignore(Deadcodeelim.elim_dead_code fd);
  S.time "loop inv check move" loopInvCheckMover fd;
  ignore(visitCilFunction checkMoverCleanup fd);
  S.time "elim_dup_checks" elim_dup_checks fd

let deadCodeElim (f:file) : unit =
  Cfg.clearFileCFG f; 
  Cfg.computeFileCFG f;
  Deadcodeelim.dce f;
  List.iter
    (fun g ->
       match g with
       | GFun (fd, _) ->
           recomputeCfg fd;
           S.time "forward substitution" forwardTmpSub fd
       | _ -> ())
    f.globals;
  Deadcodeelim.dce f;
  List.iter
    (fun g ->
       match g with
       | GFun (fd, _) -> RCT.rm_unused_locals fd
       | _ -> ())
    f.globals;
  ()

let optimFunction (fd: fundec) (l: location) =
  currentLoc := l;
  if !optLevel = 1 then
    Stats.time "optimizations" optLevel1 fd
  else if !optLevel = 2 then
    Stats.time "optimizations" optLevel2 fd
  else if !optLevel = 3 then
    Stats.time "optimizations" optLevel3 fd
