(*
 *
 * Copyright (c) 2004, 
 *  Jeremy Condit       <jcondit@cs.berkeley.edu>
 *  George C. Necula    <necula@cs.berkeley.edu>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

open Cil
open Pretty
open Doptions


module E = Errormsg
module GA = GrowArray

let curFunc : fundec ref = ref dummyFunDec
let curStmt : int ref = ref (-1)

(**************************************************************************)
(* Error/log messages *)

let d_thisloc () : doc = d_loc () !currentLoc

let bug (fmt : ('a,unit,doc,unit) format4) : 'a = 
  let f d =  
    E.hadErrors := true;
    ignore (eprintf "@!%t: Internal error: %a@!Please tell the Deputy team about this error.@!@!" d_thisloc insert d);
    flush !E.logChannel
  in
  Pretty.gprintf f fmt

let error (fmt : ('a,unit,doc,unit) format4) : 'a = 
  let f d =
    E.hadErrors := true;
    ignore (eprintf "@!%t: Error: %a@!" d_thisloc insert d);
    flush !E.logChannel
  in
  Pretty.gprintf f fmt

let unimp (fmt : ('a,unit,doc,unit) format4) : 'a = 
  let f d =
    E.hadErrors := true;
    ignore (eprintf "@!%t: Unimplemented: %a@!" d_thisloc insert d);
    flush !E.logChannel
  in
  Pretty.gprintf f fmt

let warn (fmt : ('a,unit,doc,unit) format4) : 'a = 
  let f d =
    ignore (eprintf "%t: Warning: %a@!" d_thisloc insert d);
    flush !E.logChannel
  in
  Pretty.gprintf f fmt

let log (fmt : ('a,unit,doc,unit) format4) : 'a = 
  let f d =
    ignore (eprintf "%t: %a@!" d_thisloc insert d);
    flush !E.logChannel
  in
  Pretty.gprintf f fmt

let errorwarn (fmt : ('a,unit,doc,unit) format4) : 'a = 
  let f d =
    if !trustAll then
      warn "%a" insert d
    else
      error "%a" insert d
  in
  Pretty.gprintf f fmt

(**************************************************************************)
(* Miscellaneous utility functions *)

let isPointer (e: exp) : bool =
  isPointerType (typeOf e)

let isPtrOrArray (t: typ) : bool =
  match unrollType t with
  | TPtr (bt, _)
  | TArray (bt, _, _) -> true
  | _ -> false

let isUnionType (t: typ) : bool =
  match unrollType t with
  | TComp (ci, _) when not ci.cstruct -> true
  | _ -> false

let isOpenArray (t: typ) : bool =
  match unrollType t with
  | TArray (_, None, _) -> true
  | TArray (_, Some e, _) -> isZero e
  | _ -> false

let isOpenArrayComp (t: typ) : bool =
  match unrollType t with
  | TComp (ci, _) when List.length ci.cfields > 0 ->
      let last = List.nth ci.cfields (List.length ci.cfields - 1) in
      isOpenArray last.ftype
  | _ -> false

let rec typeContainsPointers (t: typ) : bool =
  match t with
  | TPtr _
  | TFun _
  | TBuiltin_va_list _ -> true
  | TVoid _
  | TInt _
  | TFloat _
  | TEnum _ -> false
  | TArray (bt, _, _) -> typeContainsPointers bt
  | TNamed (ti, _) -> typeContainsPointers ti.ttype
  | TComp (ci, _) ->
      not ci.cdefined  (* If we don't know the contents of the struct,
                          conservatively assume it has pointers. *)
      ||
      (List.exists typeContainsPointers
         (List.map (fun fld -> fld.ftype) ci.cfields))

(** We test if a value of the given type contains as an element a pointer or 
 * a nullterminated array. This is used to decide whether to initialize with 
 * 0 an element *)
let typeContainsPtrOrNullterm (t: typ) : bool = 
  existsType 
    (function 
        TPtr _ -> ExistsTrue
      | TBuiltin_va_list _ -> ExistsTrue
      | TArray (_, _, a) when hasAttribute "nullterm" a -> ExistsTrue
      | _ -> ExistsMaybe)
    t
          
(* Is this an undefined struct? *)
let isAbstractType (t: typ) : bool =
  match unrollType t with
    | TComp (ci, _) when ci.cstruct && not ci.cdefined -> true
    | _ -> false

(* Is this a pointer to an undefined struct? *)
let isAbstractPtr (t: typ) : bool =
  match unrollType t with
    | TPtr(bt, _) -> isAbstractType bt
    | _ -> false

(* Return the base type of a pointer or array*)
let baseType (where: string) (t: typ) : typ =
  match unrollType t with
  | TPtr (bt, _)
  | TArray (bt, _, _) -> bt
  | _ -> E.s (bug "Expected pointer/array type for %s, got %a" where d_type t)

(* Return the size of the base type of a pointer or array, in bytes *)
let baseSize (t: typ) : int=
  try
    bitsSizeOf (baseType "(sizeof base type)" t) / 8
  with
  | Cil.SizeOfError (s, _) -> E.s (bug "Computed size of %s" s)
  | e -> raise e

(* Be careful when converting int64 to int.  Int64.to_int
   treats 2^31 the same as 0 *)
let to_int (i: int64) : int = 
  let i': int = Int64.to_int i in (* i.e. i' = i mod 2^31 *)
  if i = Int64.of_int i' then i'
  else E.s (E.error "Int constant too large: %Ld\n" i)

(* Interpret this integer as a signed 32-bit number.  This means we'll
   convert 0xFFFFFFFF to -1 instead of 4294967295. *)
let to_signedint (i:int64) : int =
  let (<=%) = (fun x y -> (Int64.compare x y) <= 0) in
  if -0x80000000L <=% i && i <=% 0xffffffffL then begin
    (* If 0x80000000 <= i <= 0xFFFFFFFF, subtract 2^32 from i. *)
    if 0x80000000L <=% i then
      Int64.to_int (Int64.sub i 0x1_0000_0000L)
    else
      Int64.to_int i
  end
  else
    E.s (E.error "Int constant too large: %Ld\n" i)

(* Like iter, but passes an int indicating the current index in the list. *)
let iter_index (fn: 'a -> int -> unit) (a: 'a list) : unit =
  let rec helper a n =
    match a with
    | [] -> ()
    | a1 :: arest -> fn a1 n; helper arest (n + 1)
  in
  helper a 1

(* Like iter2, but passes an int indicating the current index in the lists. *)
let iter2_index (fn: 'a -> 'b -> int -> unit) (a: 'a list) (b: 'b list) : unit =
  let rec helper a b n =
    match a, b with
    | [], [] -> ()
    | a1 :: arest, b1 :: brest -> fn a1 b1 n; helper arest brest (n + 1)
    | _ -> raise (Invalid_argument "iter2index")
  in
  helper a b 1

(* Removes the last element of a list. *)
let rec remove_last (a: 'a list) : 'a list * 'a =
  match a with
  | [] -> raise (Failure "remove_last")
  | [x] -> [], x
  | x :: rest ->
      let y, last = remove_last rest in
      x :: y, last

(**************************************************************************)
(* Expression/type comparison *)

let rec compareExp (e1: exp) (e2: exp) : bool =
(*   log "CompareExp %a and %a.\n" d_plainexp e1 d_plainexp e2; *)
  e1 == e2 ||
  match e1, e2 with
  | Lval lv1, Lval lv2
  | StartOf lv1, StartOf lv2
  | AddrOf lv1, AddrOf lv2 -> compareLval lv1 lv2
  | BinOp(bop1, l1, r1, _), BinOp(bop2, l2, r2, _) -> 
      bop1 = bop2 && compareExp l1 l2 && compareExp r1 r2
  | _ -> begin
      match isInteger (constFold true e1), isInteger (constFold true e2) with
        Some i1, Some i2 -> i1 = i2
      | _ -> false
    end

and compareLval (lv1: lval) (lv2: lval) : bool =
  let rec compareOffset (off1: offset) (off2: offset) : bool =
    match off1, off2 with
    | Field (fld1, off1'), Field (fld2, off2') ->
        fld1 == fld2 && compareOffset off1' off2'
    | Index (e1, off1'), Index (e2, off2') ->
        compareExp e1 e2 && compareOffset off1' off2'
    | NoOffset, NoOffset -> true
    | _ -> false
  in
  lv1 == lv2 ||
  match lv1, lv2 with
  | (Var vi1, off1), (Var vi2, off2) ->
      vi1 == vi2 && compareOffset off1 off2
  | (Mem e1, off1), (Mem e2, off2) ->
      compareExp e1 e2 && compareOffset off1 off2
  | _ -> false

(* Remove casts that do not effect the value of the expression, such
 * as casts between different pointer types.  Of course, these casts
 * change the type, so don't use this within e.g. an arithmetic
 * expression.
 *
 * We remove casts from pointer types to unsigned int or unsigned long.
 * 
 * We also prune casts between equivalent integer types, such as a
 * difference in sign or int vs long.  But we keep other arithmetic casts,
 * since they actually change the value of the expression. *)
let rec stripNopCasts (e:exp): exp =
  match e with
    CastE(t, e') -> begin
      match unrollType (typeOf e'), unrollType t  with
        TPtr _, TPtr _ -> (* okay to strip *)
          stripNopCasts e'
      (* strip casts from pointers to unsigned int/long*)
      | (TPtr _ as t1), (TInt(ik,_) as t2) 
          when bitsSizeOf t1 = bitsSizeOf t2 
            && not (isSigned ik) ->
          stripNopCasts e'
      | (TInt _ as t1), (TInt _ as t2) 
          when bitsSizeOf t1 = bitsSizeOf t2 -> (* Okay to strip.*)
          stripNopCasts e'
      |  _ -> e
    end
  | _ -> e
      
let compareExpStripCasts (e1: exp) (e2: exp) : bool =
  compareExp (stripNopCasts e1) (stripNopCasts e2)

(* A more conservative form of stripNopCasts.  Here, we only strip pointer
   casts if the base types have the same width.  Using this on the left operand
   of pointer arithmetic shouldn't change the resulting value. *)
let rec stripCastsForPtrArith (e:exp): exp =
  match e with
    CastE(t, e') -> begin
      match unrollType (typeOf e'), unrollType t with
        TPtr (bt1, _), TPtr (bt2, _) -> begin
          try
            if bitsSizeOf bt1 = bitsSizeOf bt2 then (* Okay to strip *)
              stripCastsForPtrArith e'
            else 
              e
          with SizeOfError _ -> (* bt1 or bt2 is abstract; don't strip. *)
            e
        end
      (* strip casts from pointers to unsigned int/long*)
      | (TPtr _ as t1), (TInt(ik,_) as t2) 
          when bitsSizeOf t1 = bitsSizeOf t2 
            && not (isSigned ik) ->
          stripCastsForPtrArith e'
      | (TInt _ as t1), (TInt _ as t2) 
          when bitsSizeOf t1 = bitsSizeOf t2 -> (* Okay to strip.*)
          stripCastsForPtrArith e'
      |  _ -> e
    end
  | _ -> e
      
let compareTypes ?(ignoreSign=true) (t1 : typ) (t2 : typ) : bool =
  let typeSigNC (t : typ) : typsig =
    let attrFilter (attr : attribute) : bool =
      match attr with
      | Attr ("poly", _) (* TODO: hack hack! *)
      | Attr ("assumeconst", _)
      | Attr ("_ptrnode", _)
      | Attr ("missing_annot", _)
      | Attr ("const", [])
      | Attr ("aligned", _)
      | Attr ("volatile", [])
      | Attr ("deprecated", [])
      | Attr ("always_inline", []) -> false
      | _ -> true
    in
    typeSigWithAttrs ~ignoreSign (List.filter attrFilter) t
  in
  (typeSigNC t1) = (typeSigNC t2)


(** return a list of vars referenced by an expression, and a boolean
  saying whether memory is referenced. *)


(**************************************************************************)

(** We remember whether the expressions that we track 
    during optimization involve a memory reference, 
    and what local vars they contain.  This makes it easier to kill facts. *)
type referenced = {
  varsRead: varinfo list;
  memRead: bool;
}

let varsOfExp (e:exp) : referenced =
  let readsVars = ref [] in
  let readsMem = ref false in
  let rec loop : exp -> unit = function
      BinOp(_, e1, e2, _) -> loop e1; loop e2
    | UnOp (_, e1, _) -> loop e1
    | CastE(_, e1) -> loop e1
    | Const _
    | SizeOf _ | SizeOfE _ | SizeOfStr _
    | AlignOf _ | AlignOfE _ -> ()
    | StartOf lv | AddrOf lv | Lval lv -> 
        let lh, off = lv in
        loopLh lh;
        loopOff off
  and loopLh: lhost -> unit =function
      Mem _ -> readsMem := true
    | Var vi -> if not (List.memq vi !readsVars) then
        readsVars := vi::!readsVars
  and loopOff: offset -> unit =function
      NoOffset -> ()
    | Field(_, off) -> loopOff off
    | Index(e, off) -> loop e; loopOff off
  in
  loop e;
  {varsRead = !readsVars; memRead = !readsMem}

let d_referenced () (r:referenced) : doc =
  dprintf "vars: %a.  mem: %s"
    (docList (fun vi -> text vi.vname)) r.varsRead
    (if r.memRead then "true" else "false")

(**************************************************************************)
(* Checks *)

type check =
    CNonNull of exp      (** e != 0 *)
  | CEq of exp * exp     (** e1 == e2 *)
  | CMult of exp * exp   (** e1 * k == e2 for some int k *)
  | CPtrArith of exp * exp * exp * exp * int
                         (** e3 + (e4 * size) does not overflow, and
                           * e1 <= e3 + (e4 * size) <= e2. *)
  | CPtrArithNT of exp * exp * exp * exp * int
                         (** e3 + (e4 * size) does not overflow, and
                           * e1 <= e3 + (e4 * size) <= (e2 + sizeof(e2)). *)
  | CLessInt of exp * exp * string
                         (** e1 < e2, unsigned.
                           * Also remember why this check was added. *)
  | CLeqInt of exp * exp * string
                         (** e1 <= e2, unsigned.
                           * Also remember why this check was added. *)
  | CLeq of exp * exp * string
                         (** e1 <= e2, unsigned.
                           * Also remember why this check was added. *)
  | CLeqNT of exp * exp * int * string
                         (** e1 <= (e2 + sizeof(e2)), unsigned.
                           * The int is the size of the base type.
                           * Also remember why this check was added. *)
  | CNullOrLeq of exp * exp * exp * string
                         (** e1 == 0 || e2 <= e3.
                           * Also remember why this check was added. *)
  | CNullOrLeqNT of exp * exp * exp * int * string
                         (** e1 == 0 || e2 <= (e3 + sizeof(e3)).
                           * The int is the size of the base type.
                           * Also remember why this check was added. *)
  | CWriteNT of exp * exp * exp * int
                         (** (e1 == e2) ==> (e3 = 0)
                           * The int is the size of the base type. *)
  | CNullUnionOrSelected of lval * exp
                         (** lv = \vec{0} || e.
                             Here, e is a shortcut saying that if the 
                             newly-active field is the same as the old
                             active field, we don't check for the union 
                             equalling zero.*)
  (* These two are redundant with CNonNull and CEq, but having separate
     checks for unions gives better error messages: *)
  | CSelected of exp     (** e != 0 *)
  | CNotSelected of exp  (** e == 0 *)
(* Other checks will be needed, such as nullterm checks and checks for when
   part of one of the above checks can be proved statically. *)

let checks_equal c1 c2 = match c1, c2 with
| CEq(e11,e12), CEq(e21,e22)
| CMult(e11,e12), CMult(e21, e22)
| CLessInt(e11,e12,_), CLessInt(e21,e22,_)
| CLeqInt(e11,e12,_), CLeqInt(e21,e22,_)
| CLeq(e11,e12,_), CLeq(e21,e22,_) ->
    (compareExpStripCasts e11 e21) &&
    (compareExpStripCasts e12 e22)
| CLeqNT(e11,e12,sz1,_), CLeqNT(e21,e22,sz2,_) ->
    (compareExpStripCasts e11 e21) &&
    (compareExpStripCasts e12 e22) &&
    sz1 = sz2
| CNullOrLeq(e11,e12,e13,_), CNullOrLeq(e21,e22,e23,_) ->
    (compareExpStripCasts e11 e21) &&
    (compareExpStripCasts e12 e22) &&
    (compareExpStripCasts e13 e23)
| CNullOrLeqNT(e11,e12,e13,sz1,_), CNullOrLeqNT(e21,e22,e23,sz2,_)
| CWriteNT(e11,e12,e13,sz1),CWriteNT(e21,e22,e23,sz2) ->
    (compareExpStripCasts e11 e21) &&
    (compareExpStripCasts e12 e22) &&
    (compareExpStripCasts e13 e23) &&
    sz1 = sz2
| CPtrArith(e11,e12,e13,e14,sz1), CPtrArith(e21,e22,e23,e24,sz2)
| CPtrArithNT(e11,e12,e13,e14,sz1), CPtrArithNT(e21,e22,e23,e24,sz2) ->
    (compareExpStripCasts e11 e21) &&
    (compareExpStripCasts e12 e22) &&
    (compareExpStripCasts e13 e23) &&
    (compareExpStripCasts e14 e24) &&
    sz1 = sz2
| CNullUnionOrSelected(l1, e1), CNullUnionOrSelected(l2, e2) ->
    (compareLval l1 l2) &&
    (compareExpStripCasts e1 e2)
| CSelected e1, CSelected e2
| CNotSelected e1, CNotSelected e2
| CNonNull e1, CNonNull e2 ->
    compareExpStripCasts e1 e2
| _ -> false

(* These aren't real variables.  In the output, they'll show up as
   __LOCATION__, which is a macro defined in deputy/checks.h. We use them
   for calling runtime check functions. *)
let locationToken : exp=
  let vi = makeGlobalVar "__LOCATION__" charPtrType in
  Lval (var vi)

let mkFun (name: string) (rt:typ) (args: typ list) : exp = 
  let fdec = emptyFunction name in
  let args = List.map (fun t -> ("", t, [])) args in
  fdec.svar.vtype <- TFun(rt, Some args, false, []);
  Lval (var fdec.svar)

let mkCheckFun (name: string) (n: int) : exp = 
 (* A check function takes n void* parameters, a location *)
  let args = Util.list_init n (fun _ -> voidPtrType) in
  let args' = args @ [charPtrType] in
  mkFun name voidType args'

let cnonnull = mkCheckFun "CNonNull" 1
let ceq = mkCheckFun "CEq" 2
let cmult = mkCheckFun "CMult" 2
let cptrarith = mkCheckFun "CPtrArith" 5
let cptrarithnt = mkCheckFun "CPtrArithNT" 5
let clessint = mkCheckFun "CLessInt" 3
let cleqint = mkCheckFun "CLeqInt" 3
let cleq = mkCheckFun "CLeq" 3
let cleqnt = mkCheckFun "CLeqNT" 4
let cnullorleq = mkCheckFun "CNullOrLeq" 4
let cnullorleqnt = mkCheckFun "CNullOrLeqNT" 5
let cwritent = mkCheckFun "CWriteNT" 4
let cnullunion = mkCheckFun "CNullUnionOrSelected" 3
let cselected = mkCheckFun "CSelected" 1
let cnotselected = mkCheckFun "CNotSelected" 1

let unmkString (e: exp) : string =
  match e with
  | Const (CStr s) -> s
  | _ -> E.s (bug "Expected string constant")

let toInt (e:exp) : int =
  match isInteger e with
    Some i64 -> to_int i64
  | None -> 
      E.s (bug "expected a constant int for the size param in this check.")

let checkFunctions : exp list =
  [ cnonnull; ceq; cmult; cptrarith; cptrarithnt; clessint; cleqint;
    cleq; cleqnt; cnullorleq; cnullorleqnt; cwritent; cnullunion;
    cselected; cnotselected ]

let instrToCheck (instr: instr) : check option =
  match instr with
  | Call (None, fn, args, _) when List.exists (compareExp fn) checkFunctions ->
      let c =
        match args with
        | [e;_] when compareExp fn cnonnull ->
            CNonNull e
        | [e1;e2;_] when compareExp fn ceq ->
            CEq (e1,e2)
        | [e1;e2;_] when compareExp fn cmult ->
            CMult (e1,e2)
        | [e1;e2;e3;e4;e5;_] when compareExp fn cptrarith ->
            CPtrArith (e1,e2,e3,e4,toInt e5)
        | [e1;e2;e3;e4;e5;_] when compareExp fn cptrarithnt ->
            CPtrArithNT (e1,e2,e3,e4,toInt e5)
        | [e1;e2;why;_] when compareExp fn clessint ->
            CLessInt (e1,e2,unmkString why)
        | [e1;e2;why;_] when compareExp fn cleqint ->
            CLeqInt (e1,e2,unmkString why)
        | [e1;e2;why;_] when compareExp fn cleq ->
            CLeq (e1,e2,unmkString why)
        | [e1;e2;e3;why;_] when compareExp fn cleqnt ->
            CLeqNT (e1,e2,toInt e3,unmkString why)
        | [e1;e2;e3;why;_] when compareExp fn cnullorleq ->
            CNullOrLeq (e1,e2,e3,unmkString why)
        | [e1;e2;e3;e4;why;_] when compareExp fn cnullorleqnt ->
            CNullOrLeqNT (e1,e2,e3,toInt e4,unmkString why)
        | [p;hi;what;sz;_] when compareExp fn cwritent ->
            CWriteNT (p,hi,what,toInt sz)
        | [AddrOf lv;_;e;_] when compareExp fn cnullunion ->
            CNullUnionOrSelected (lv, e)
        | [e;_] when compareExp fn cselected ->
            CSelected (e)
        | [e;_] when compareExp fn cnotselected ->
            CNotSelected (e)
        | _ ->
            E.s (bug "Check instruction not recognized: %a" d_instr instr)
      in
      Some c
  | _ -> None

let checkToInstr (c:check) : instr =
  let call f args = Call(None, f,
                         (* Append the file and line to the end of the args *)
                         args @ [locationToken ],
                         !currentLoc) 
  in
  let i =
    match c with
      CNonNull (e) -> call cnonnull [e]
    | CEq (e1,e2) -> call ceq [e1;e2]
    | CMult (e1,e2) -> call cmult [e1;e2]
    | CPtrArith (e1,e2,e3,e4,e5) -> call cptrarith [e1;e2;e3;e4;integer e5]
    | CPtrArithNT (e1,e2,e3,e4,e5) -> call cptrarithnt [e1;e2;e3;e4;integer e5]
    | CLessInt (e1,e2,why) -> call clessint [e1;e2;mkString why]
    | CLeqInt (e1,e2,why) -> call cleqint [e1;e2;mkString why]
    | CLeq (e1,e2,why) -> call cleq [e1;e2;mkString why]
    | CLeqNT (e1,e2,e3,why) -> call cleqnt [e1;e2;integer e3;mkString why]
    | CNullOrLeq (e1,e2,e3,why) -> call cnullorleq [e1;e2;e3;mkString why]
    | CNullOrLeqNT (e1,e2,e3,e4,why) ->
        call cnullorleqnt [e1;e2;e3;integer e4;mkString why]
    | CWriteNT (p,hi,what,sz) -> call cwritent [p;hi;what;integer sz]
    | CNullUnionOrSelected (lv, sameFieldSelected) -> 
        let sz = sizeOf (typeOfLval lv) in
        call cnullunion [mkAddrOf lv; sz; sameFieldSelected]
    | CSelected (e) -> call cselected [e]
    | CNotSelected (e) -> call cnotselected [e]
  in
  (* For the optimizer to work properly, we must be able to convert instrs
   * back to the original check.  As a sanity check, we verify here that
   * this is possible for each instr we generate. *)
  if instrToCheck i = None then
    E.s (bug "checkToInstr not invertible");
  i

(*
 * Return true if i is a deputy
 * runtime check.
 *)
let is_check_instr i =
  match instrToCheck i with
    None -> false
  | Some _ -> true

(* 
 * Return true if i is any
 * instruction that deputy added.
 *)
let is_deputy_instr i =
  match instrToCheck i with
    Some _ -> true
  | None -> match i with
      Call(_,Lval(Var vf,NoOffset),_,_) ->
	vf.vname = "deputy_findnull" ||
	vf.vname = "deputy_max"
      | _ -> false

let rec expRefersToVar (name: string) (e: exp) : bool =
  match e with
  | Lval lv -> lvalRefersToVar name lv
  | AddrOf lv -> lvalRefersToVar name lv
  | StartOf lv -> lvalRefersToVar name lv
  | SizeOfE e' -> expRefersToVar name e'
  | AlignOfE e' -> expRefersToVar name e'
  | UnOp (_, e', _) -> expRefersToVar name e'
  | BinOp (_, e1, e2, _) -> expRefersToVar name e1 || expRefersToVar name e2
  | CastE (_, e') -> expRefersToVar name e'
  | Const _
  | SizeOf _
  | SizeOfStr _
  | AlignOf _ -> false

and lvalRefersToVar (name: string) ((host, offset): lval) : bool =
  let rec offsetRefersToVar (offset: offset) =
    match offset with
    | Field (fld, offset') -> offsetRefersToVar offset'
    | Index (e, offset') -> expRefersToVar name e || offsetRefersToVar offset'
    | NoOffset -> false
  in
  match host with
  | Var vi -> vi.vname = name || offsetRefersToVar offset
  | Mem e -> expRefersToVar name e || offsetRefersToVar offset


(* CIL treats these functions specially and uses & in the arguments
   lists for its own reasons.*)
let isVarargOperator (f:exp) : bool =
  let varargOperators = [
    "__builtin_va_arg";
    "__builtin_stdarg_start";
    "__builtin_va_start";
    "__builtin_next_arg";
  ]
  in
  match f with
  | Lval(Var vi, NoOffset) -> List.mem vi.vname varargOperators
  | _ -> false

(* Is "size" the correct size of the given lval?
   We use this when checking calls to memset. *)
let isCorrectSizeOf (size: exp) (lv: lval) : bool =
  (* return true if size is an expression for the size of lv. *)
  let actualSize : int = (bitsSizeOf (typeOfLval lv)) / 8 in
  let size' : int64 option = isInteger (constFold true size) in
  size' = Some (Int64.of_int actualSize)


let memset: varinfo =
    (* Create a decl for "__deputy_memset". This is identical to memset,
       but we give it its own name to avoid conflicts. We add this to the file
       in dinfer. *)
    let memsetType = TFun(voidPtrType, Some [("p", voidPtrType, []);
                                             ("what", intType, []);
                                             ("sz", ulongType, [])],
                          false,
                          [Attr("dmemset", [AInt 1; AInt 2; AInt 3])]) in
    makeGlobalVar "__deputy_memset" memsetType;
