(*
 *
 * Copyright (c) 2001-2006, 
 *  Matt Harren        <matth@cs.berkeley.edu>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

open Cil
open Pretty
open Dutil
open Dattrs

module E = Errormsg
module IH = Inthash
module H = Hashtbl

module N = Ptrnode
module MU = Markutil

let missingAnnotation = ref false

(* similar to examineType, but for listGlobalAnnotations instead of
   printGlobalAnnotations. *)
let rec examineNode (* ?(suppressBoundsWarning=false) *) 
  (where: string) (t:typ): unit =
  let reportNodeDoc (what:doc): unit =
    missingAnnotation := true;
    log "@[node \"%a\"\nin %s %a@]" dx_type t where insert what
  in
  let reportNode (what:string): unit =
    reportNodeDoc (text what)
  in
  let t = unrollType t in
  let nodeOf (a:attributes) : N.node =
    match  N.nodeOfAttrlist a with
      None -> E.s (bug "Missing node for %a at %s" d_type t where)
    | Some n -> n
  in
  match t with
    TVoid _
  | TBuiltin_va_list _
  | TInt _
  | TFloat _
  | TEnum _ -> ()

  | TPtr (bt, a)
  | TArray (bt, _, a) when hasAttribute "trusted" a ->
      (* listExamineType where bt; *)
      ()

  | TPtr (bt, a) ->
      examineNode where bt;
      let k, _ = N.inferredKindOf a in
      (* log "Considering %a with kind %a." dx_type t N.d_opointerkind k; *)
      let needsBounds =
        N.kindNeedsBounds k &&
        not (hasAttribute "bounds" a) &&
        not (hasAttribute "size" a)
      in
      let needsNT =
        N.kindIsNullterm k &&
        not (hasAttribute "nullterm" a)
      in
      (match needsBounds, needsNT with
         true, true -> reportNode "should be annotated NTS (or NT COUNT(...), if possible)."
       | true, false -> reportNode "needs a bound annotation."
       | false, true -> reportNode "should be annotated NT."
       | false, false -> ()
      );
      if isVoidType bt && not (hasAttribute "size" a) then begin
        let n = nodeOf a in
        try
          let other,loc = List.assq n !MU.voidStarCasts in
          reportNodeDoc 
            (dprintf "is a void* that is cast to/from type %a at %a"
               d_type (TPtr(other.N.btype,other.N.attr)) d_loc loc)
        with Not_found -> ()
      end;
(*       if N.hasFlag n N.pkStack && not (hasAttribute "noescape" a) then begin *)
(*         reportNode n "points to the stack" *)
(*       end; *)
      ()
  | TArray (bt, lo, a) ->
      examineNode where bt;
      let k, _ = N.inferredKindOf a in
      if N.kindIsNullterm k && not (hasAttribute "nullterm" a) then
        reportNode "should be annotated NT.";
      if isOpenArray t && not (hasAttribute "bounds" a) then begin
        reportNode "contains an open array"
      end
  | TFun(rt, args, isva, a) ->
(*       if isva then begin *)
(*         incr ctr.vararg; *)
(*         doPrintInd (text "Vararg function.\n") *)
(*       end; *)
      examineNode "<function pointer return type>" rt;
      List.iter (fun (n, t, _) -> 
                   examineNode ("<function pointer formal "^n^">") t) 
        (argsToList args)
  | TComp _ -> ()(* We visit these separately. *)
  | TNamed _ -> E.s (bug "unrollType")


let reportNeededAnnotations (f:file) : unit =
  missingAnnotation := false;
  E.log "\nChecking whether additional annotations are needed for Deputy...\n";
  let oldPCI = !print_CIL_Input in
  print_CIL_Input := true;
  
  let totKind : (N.opointerkind, int ref) H.t = H.create 17 in
  let totalNodes : int ref = ref 0 in
  (* Collect stats on the kinds of every local or cast that might
     need inferred bounds *)
  let visitLocalOrCast (t:typ): unit =
    match unrollType t with
      TPtr(_,a) ->
        incr totalNodes;
        let k,_ = N.inferredKindOf a in
        N.addToHisto totKind 1 k
    | _ -> ()
  in
  let visitAllCasts (fd: fundec) : unit =
    let castVisitor = object (self) 
      inherit nopCilVisitor
      method vexpr e =
        (match e with
           CastE(t, _) -> visitLocalOrCast t
         | _ -> ());
        DoChildren
    end
    in
    ignore(visitCilBlock castVisitor fd.sbody)
  in

  (* Maintain a map from globals to the location of their first declaration.
     If a var has no location (probably because it's Poly), use the 
     location of it's declaration.  *)
  let declarations: location IH.t = IH.create 50 in
  let setLoc vi: unit =
    if !currentLoc == locUnknown then begin
      if IH.mem declarations vi.vid then
        currentLoc := IH.find declarations vi.vid
    end else
      if not (IH.mem declarations vi.vid) then
        IH.add declarations vi.vid !currentLoc
  in
  let globalsVisited: unit IH.t = IH.create 50 in
  let doGlobal g: unit =
    match g with
      GVarDecl (vi, l) when Dattrs.isSpecialFunction vi.vtype ->
        ()
    | GVarDecl (vi, l) when isFunctionType vi.vtype 
        && not (IH.mem globalsVisited vi.vid) ->
        currentLoc := l;
        setLoc vi;
        (* E.log "visiting %s at %a\n" vi.vname d_loc l; *)
(*         if Wrappers.hasAWrapper (Poly.stripPoly vi.vname) then *)
(*           E.log "%a: @[extern function %s has a wrapper,\n   so some needed annotations may not be shown@]\n" d_loc l vi.vname; *)
          let rt, formals, _, _ = splitFunctionType vi.vtype in
          examineNode ("the return value of "^vi.vname) rt;
          List.iter (fun (n,t,_) -> 
                       examineNode ("formal \""^n^"\" of "^vi.vname) t)
            (argsToList formals);
        IH.add globalsVisited vi.vid ()

    | GFun (fdec, l) when Dattrs.isTrustedType fdec.svar.vtype ->
        ()
    | GFun (fdec, l) ->
        currentLoc := l;
        setLoc fdec.svar;
        (* E.log "visiting %s at %a\n" fdec.svar.vname d_loc l; *)
        if not (IH.mem globalsVisited fdec.svar.vid) then begin
          (* Examine the prototype, since we haven't already seen a
             declaration of this func: *)
          IH.add globalsVisited fdec.svar.vid ();
          let rt, formals, _, _ = splitFunctionType fdec.svar.vtype in
          examineNode ("the return value of "^fdec.svar.vname) rt;
          List.iter (fun vi -> examineNode 
                       ("formal \""^vi.vname^"\" of "^fdec.svar.vname)
                       vi.vtype)
            fdec.sformals;
        end;
        (* Regardless of whether we've seen an earlier declaration, examine
           locals and casts. *)
        (* Locals are mostly covered by inference.  But base types of
           pointers may need annotating. *)
        List.iter (fun vi -> 
                     match unrollType vi.vtype with 
                       TPtr (bt, a) as t-> 
                         visitLocalOrCast t;
                         examineNode ("local \""^vi.vname^"\"") bt
                     | TArray (bt, _, a) -> 
                         visitLocalOrCast bt;
                         examineNode ("local \""^vi.vname^"\"") bt
                     | _ -> ())
          fdec.slocals;
        visitAllCasts fdec

    | GVarDecl (vi, l)
    | GVar (vi, _, l) when not (IH.mem globalsVisited vi.vid) ->
        currentLoc := l;
        setLoc vi;
        IH.add globalsVisited vi.vid ();
        examineNode ("global \""^vi.vname^"\"") vi.vtype
    | GCompTag (ci, l) ->
        currentLoc := l;
        List.iter
          (fun fi -> examineNode ("field \""^fi.fname^"\"") fi.ftype)
          ci.cfields
    | _ -> ()
  in
  iterGlobals f doGlobal;
  if !missingAnnotation then
    E.log "Done listing annotations needed for Deputy.\n\n"
  else
    E.log "All globals and fields are annotated.\n";

  let percent (n: int) : float = 
    if !totalNodes = 0 then begin
      100.0
    end else
      (float_of_int n)
         /. float_of_int (!totalNodes) *. 100.0 
  in
  E.log "Inferred pointer kinds for %d locals and casts:\n" !totalNodes;
  H.iter
    (fun k r -> ignore (E.log "    %a - %d (%3.0f%%)\n"
                          N.d_opointerkind k !r (percent !r)))
    totKind;
  E.log "\n";
  print_CIL_Input := oldPCI;
  ()

(* The entry point from Deputy.  Calls Markptr, Solver *)
let inferKinds (f:file) : file = 
  if f.globinit <> None then
    (* CCured has special handling of globinit, but we can't use that
       strategy because main may not be in this file. *)
    ignore (E.warn "Inference: skipping global initializers.\n");
  (* if !verbose then E.log "Interprocedural inference: Calling Markptr\n"; *)
  let marked = Stats.time "markptr" Markptr.markFile f in

  Util.tryFinally
    (fun _ ->
       Stats.time "solver"
       (fun graph ->
          (* Here we actually call the solver *)
          let should_typecheck = Solver.solve marked graph in 
          if should_typecheck then begin 
            ignore (E.warn "The inference results should be type-checked.")
          end ; 
       ) Ptrnode.idNode
    )
    (* Now the finally clause *)
    (fun _ ->
(*        if !doCure then  *)
(*           Stats.time "graphstats" Ptrnode.printGraphStats (); *)
       (match !Ptrnode.browserFile with
          None -> ()
        | Some(f)->
            Stats.time "printBrowser" (Ptrnode.printBrowser f) marked))
    ();

  reportNeededAnnotations marked;
  MU.voidStarCasts := [];

  marked
