(*
 *
 * Copyright (c) 2006,
 *  George C. Necula    <necula@cs.berkeley.edu>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

module F = Frontc
module C = Cil
module E = Errormsg


type outfile = { fname: string; fchan: out_channel } 
let outChannel : outfile option ref = ref None

exception Done_Processing

let printFile ~(extraPrinting:(out_channel->unit) option) 
              ~(globinit:C.fundec option)
             (f: C.file) (name: string) : unit
  =
  if name <> "" then begin
    try
      let channel = open_out name in
      C.dumpFile Dattrs.deputyPrinter channel name f;
      (match globinit with 
        Some g -> C.dumpGlobal Dattrs.deputyPrinter 
                               channel (C.GFun(g, C.locUnknown))
      | None -> ()); 
      (match extraPrinting with
         Some doit -> doit channel
       | None -> ());
      close_out channel
    with Sys_error _ ->
      E.s (E.error "Error dumping inference results to %s\n" name)
  end

let parseOneFile (fname: string) : C.file =
  let cil = F.parse fname () in
  Rmtmps.removeUnusedTemps ~isRoot:(Dattrs.treatAsRoot cil) cil;
  cil

let rec processOneFile (cil: C.file) =
  try
    printFile ~extraPrinting:None ~globinit:None cil !Doptions.parseFile;
    Dinfer.preProcessFile cil;

    (* "marked" is the original file plus NODE() annotations. *)
    let marked = 
      if !Doptions.inferKinds then
        Stats.time "Interprocedural inference" Inferkinds.inferKinds cil
      else
        cil
    in
    Dinfer.preProcessFileAfterMarking marked;
    (* Now insert the function with the global initializers. Does not insert 
     * it in the file. *)
    let globinit = Globinit.prepareGlobalInitializers marked in
    
    let extraprint = 
      if !Doptions.inferKinds && !Doptions.emitGraphDetailLevel >= 0 then
        Some Ptrnode.printInferGraph
      else
        None
    in
    printFile ~extraPrinting:extraprint 
              ~globinit:(Some globinit) marked !Doptions.inferFile;

    Dcheck.checkFile marked globinit;

    (* Now do the global initializer *)
    let residualChecks: bool =
      Globinit.checkGlobinit marked globinit
        (fun gi -> Dcheck.checkFundec gi)
        (fun gi l -> 
             Doptim.optimFunction gi l;
             ignore (C.visitCilBlock Dinfer.postPassVisitor gi.C.sbody)) in

    if residualChecks then
      marked.C.globals <- marked.C.globals @ [C.GFun (globinit, C.locUnknown)];

    Dinfer.postProcessFile marked;

    begin
      match !outChannel with
      | None -> ()
      | Some c ->
          (* Tell CIL to put comments around the bounds attributes. *)
          C.print_CIL_Input := false;
          Stats.time "printCIL" 
               (C.dumpFile (!C.printerForMaincil) c.fchan c.fname) marked
    end;
    let optTime = Stats.lookupTime "optimizations" in
    if optTime >= 20.0 && !Doptions.optLevel > 1 then
      E.log("\nNote: Optimizations took %.1f s.  For faster (but less precise)"
            ^^" analysis, you can use a lower optimization level.\n\n")
        optTime;

    if !E.hadErrors then
      E.s (E.error "Error while processing file");
  with Done_Processing -> ()
        
let rec main () =
  let usageMsg = "Usage: deputy [options] source-files" in

  (* sm: enabling this by default, since I think usually we
   * want 'cilly' transformations to preserve annotations; I
   * can easily add a command-line flag if someone sometimes
   * wants these suppressed *)
  C.print_CIL_Input := true;

  (* Turn off the implicit casts *)
  C.insertImplicitCasts := false;

  (* Make TRUSTED an attribute of types, not names.
   * (This is just for the sake of uniformity) *)
  Hashtbl.add C.attributeHash "trusted" C.AttrType;

  (* Tell cabs2cil to strip dependent attributes from the types of
     temporaries that it creates and casts that it inserts. *)
  Cabs2cil.typeForInsertedVar := Dattrs.stripDepsFromType;
  Cabs2cil.typeForInsertedCast := Dattrs.stripDepsFromType;

  let openFile (takeit: outfile -> unit) (fl: string) = 
    try
      takeit { fname = fl; fchan = open_out fl }
    with _ ->
      raise (Arg.Bad ("Cannot open output file " ^ fl))
  in

  let argDescr =
    Ciloptions.options @ 
    ["--out",
     Arg.String (openFile (fun oc -> outChannel := Some oc)),
     "output file"] @
    F.args @
    ["", Arg.Unit (fun () -> ()), 
       "\n\n     ************   Deputy Options ************"] @
    Doptions.options @
    ["", Arg.Unit (fun () -> ()), "\n"]
  in

  Stats.reset false; (* no performance counters *)

  Arg.parse argDescr Ciloptions.recordFile usageMsg;
  Cil.initCIL ();

  Ciloptions.fileNames := List.rev !Ciloptions.fileNames;

  let fileName =
    match !Ciloptions.fileNames with
    | [name] -> name
    | [] -> E.s (E.error "No file names provided\n")
    | _ -> E.s (E.error "Too many file names provided (%a)\n"
               (Pretty.docList Pretty.text) !Ciloptions.fileNames)
  in

  let file = parseOneFile fileName in

  if !E.hadErrors then
    E.s (E.error "Cabs2cil had some errors");

  processOneFile file
;;

let failed = ref false 

let cleanup () = 
  if !E.verboseFlag || !Cilutil.printStats then
    Stats.print !E.logChannel "Timings:\n";
  if !E.logChannel != stderr then 
    close_out (! E.logChannel);  
  (match ! outChannel with Some c -> close_out c.fchan | _ -> ())
;;

begin 
  try 
    main () 
  with F.CabsOnly -> (* this is OK *) ()
end;
cleanup ();
exit (if !failed then 1 else 0)
