(*
 *
 * Copyright (c) 2001-2002, 
 *  George C. Necula    <necula@cs.berkeley.edu>
 *  Scott McPeak        <smcpeak@cs.berkeley.edu>
 *  Wes Weimer          <weimer@cs.berkeley.edu>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

open Cil
open Pretty
open Dattrs
open Dutil

module H = Hashtbl
module E = Errormsg

module MU = Markutil


let debugVararg = false


(* Compute the signature of a type for comparing arguments *)
let argumentTypeSig (t: typ) = 
  let rec argumentPromotion (t : typ) : typ =
    match unrollTypeDeep t with
      (* We assume that an IInt can hold even an IUShort *)
      TInt ((IShort|IUShort|IChar|ISChar|IUChar
             |ILong|IULong|IUInt|IInt), a) -> 
        TInt(IInt, [])
      (* We always drop the unsigned and the attributes on integer *)
    | TInt ((ILongLong|IULongLong), a) -> TInt(ILongLong, [])
          (* floats are passed as double *)
    | TFloat ((FFloat|FDouble), a) -> TFloat (FDouble, [])
    | TEnum (_, a) -> TInt(IInt, [])
    | TPtr(TInt((IChar|IUChar|ISChar), a1), a2) -> TPtr(TInt(IChar, []), [])
    | t -> t
  in
  (* Ignore all attributes *)
  typeSigWithAttrs (fun x -> []) (argumentPromotion t)


(** A structure for processing variable arguments **)
type vaProcessor = {
  vapName: string;    (* name of the processor *)
  vapProcessArgs: (exp list -> instr list * exp list); 
                      (* process the argument list. Return some checking 
                       * instructions and the modified argument list *)
  }


(** Fold-right over the dot-dot arguments *)
let foldDotDotArgs (processOne: exp -> 'a -> exp * 'a)
                   (skipArgs: int)
                   (args: exp list) 
                   (acc: 'a) : exp list * 'a = 
  let rec loop (skip: int) (args: exp list) (acc: 'a) : exp list * 'a = 
    match args with 
      [] -> if skip > 0 then 
               E.s (error "Too few arguments in vararg call");
            [], acc
    | a :: resta when skip > 0 -> 
       let a'l, acc' = loop (skip - 1) resta acc in
       a :: a'l, acc'

    | a :: resta -> begin
       let a'l, acc' = loop 0 resta acc in 
       let a', acc'' = processOne a acc' in 
       a' :: a'l, acc''
    end
  in 
  loop skipArgs args acc


(** Process an actual argument to a function. Return replacement actual and a 
 * string describing the argument **)
let processActualArgument 
    (arg: exp) 
    (descr: (typ * typsig * string) list)  (* a list of expected types, the 
                                            * corresponding type signature, 
                                            * and a descriptor *)
    : exp * string = 
  let argt = typeOf arg in 
  let args = argumentTypeSig argt in 
  try 
    let (t, _, d) = List.find (fun (_, s, _) -> Util.equals s args) descr in
    mkCastT arg argt t, d
  with Not_found -> 
    raise Not_found 




(** The list of format descriptors *)
let formatDescriptors : (typ * (* type of formal *)
                         typsig * (* signature of formal *)
                         string) (* descriptor of formal *)
    list = 
  List.map (fun (t, d) -> (t, argumentTypeSig t, d)) 
    [ (TPtr(charType, Dattrs.stringAttrs), "%s");
      (intType, "%d");
      (doubleType, "%f");
       (TPtr(intType, []), "%n")
    ]
    
(** Construct the processor for format strings *)
let printfProcessor (nrFormals: int) (formidx: int) : vaProcessor = 
  if formidx <= 0 || formidx > nrFormals then 
    E.s (error "Expected the format string to be the %dth argument"
           formidx);
  let rec checkFormatArgs (argidx: int) 
                          (actual: string) (aidx: int) 
                          (expected: string) (eidx: int) 
                           = 
      let a, aidx' = findNextDescr actual aidx in
      let e, eidx' = findNextDescr expected eidx in 
      if debugVararg then 
           log "  next_descr=%c (%d), next_actual=%c (%d)"
                  a aidx' e eidx';
      if a = '_' then (* The actuals are over *)
         if e = '_' then 
            ()
         else
(*            error "Too few arguments to printf-like vararg function" *)
            warn "BAD: Too few arguments to printf-like vararg function"
      else if e = '_' then (* The expected are over *)
         warn "Too many arguments to printf-like vararg function"
      else if a = e then   
         checkFormatArgs (argidx+1) actual aidx' expected eidx' 
      else 
(*         E.s (error "Argument type (no. %d) in printf-like function matches descriptor \"%%%c\" but was expected to match \"%%%c\"\n" argidx a e) *)
         warn "BAD: Argument type (no. %d) in printf-like function matches descriptor \"%%%c\" but was expected to match \"%%%c\"\n" argidx a e

  (* Return the index of the next descriptor *)
  and findNextDescr (s: string) (from: int) : char * int = begin
    try
       let i = String.index_from s from '%' in
       let rec parseConversionSpec (start: int) : char * int =
         match Char.lowercase s.[start] with
           'e' | 'f' | 'g' |   (* double arg *)
           'a' ->              (* double arg *)
            'f', start + 1
         | 'd' | 'i' |         (* signed int arg *)
           'o' | 'u' | 'x' |   (* unsigned int arg *)
           'c' |               (* unsigned char *)
           'p' ->              (* void pointer treated as int *)
            'd', start + 1
         | 's' -> 's', start + 1 (* char pointer *)
         | 'n' -> 'n', start + 1
         | _ -> parseConversionSpec (start+1)
      in
      parseConversionSpec i
    with _ ->
        '_', -1
  end in


  let doArgs (args: exp list) : instr list * exp list = 
    if List.length args < nrFormals then 
       E.s (error "Too few arguments to vararg function");
    (* Process the arguments *)
    let (args': exp list), (descr: string) = 
     foldDotDotArgs 
       (fun a acc -> 
         try
           let a', d = processActualArgument a formatDescriptors in
           a', d ^ acc
         with Not_found -> begin
           (* Cast it to integer, perhaps it is a "p" format. *)
           (* matth: don't warn about this.  casting pointers to ints to print
              with %p or %x is common, especially in the kernel. *)
           (* warn "Cast vararg argument %a to integer" d_exp a; *)
           (mkCast a intType), "%d" ^ acc
         end)
     nrFormals
     args
     ""
    in
    (* Now check the format string *)
    (match List.nth args (formidx - 1) with 
      Const (CStr form) -> 
        if debugVararg then 
           log "vararg: checkFormat(\"%s\", \"%s\")" descr form;
        checkFormatArgs nrFormals descr 0 form 0
    | _ -> 
      warn "Non-literal format string. Cannot check that it conforms with \"%s\"\n"
          descr);
    [], args'
  in
  { vapName = "printf";
    vapProcessArgs = doArgs
  }
    
  

(** Construct a vaProcessor for a given function. **)
let getVaProcessor (func: exp) : vaProcessor = 
  match unrollType (typeOf func) with
    TFun(_, Some forms, _, a) -> begin
       match filterAttributes "dvararg" a with 
          Attr(_, [ACons("printf", [AInt fidx])]) :: _ -> 
            printfProcessor (List.length forms) fidx
              
       | a -> raise Not_found
    end
  | _ -> assert false
  
      
(* Prepare the arguments in a call to a vararg function. *)
let prepareVarargArguments
    ~(mkTempVar: typ -> varinfo) (* how to make a temporary variable *)
    ~(func: exp)                 (* the called function *)
    ~(nrformals: int)            (* how many formals *)
    ~(args: exp list) : instr list * exp list = 

  if debugVararg then 
    ignore (log "Preparing args in call to %a" d_exp func);
  try
    let processor : vaProcessor = getVaProcessor func in 
    if debugVararg then 
      ignore (log "  will use the processor: %s" processor.vapName);
    
    processor.vapProcessArgs args
  with Not_found -> begin
    warn "Call to vararg function %a not fully checked\n" d_exp func;
    [], args
  end




(* Turn the gcc format attribute into our own notation.  Returns a type
   that may have a new attribute but which is otherwise the same. *)
let processFormatAttribute (funType: typ) : typ = 
  match filterAttributes "format" (typeAttrs funType)
        @ filterAttributes "__format__" (typeAttrs funType)
  with
    [Attr(_, [printf; AInt format_idx; AInt _])] 
      when (match printf with 
                ACons("printf",[])  
              | ACons("__printf__",[]) -> true 
              | _ -> false) -> begin

        match funType with 
         TFun(rt, forms, isva, a) -> 
           TFun(rt, forms, isva, 
                addAttribute (Attr("dvararg", [ACons("printf", 
                                                     [AInt format_idx])]))
                             a)
        | _ -> assert false
      end

  | [] -> funType
  | al -> ignore (warn "Do not understand %a" d_attrlist al); funType



let initFile (f: file) = ()
